# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
from collections.abc import Iterable, Sequence
from typing import Any, cast

import pytest

from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import SortMode
from astrapy.data_types import DataAPITimestamp, DataAPIVector
from astrapy.exceptions import (
    DataAPIException,
    DataAPIResponseException,
    TableInsertManyException,
)
from astrapy.results import TableInsertManyResult

from ..conftest import (
    IS_ASTRA_DB,
    DefaultTable,
    _repaint_NaNs,
    _typify_tuple,
)
from .table_row_assets import (
    AR_ROW_0,
    AR_ROW_0_B,
    AR_ROW_PK_0,
    AR_ROW_PK_0_TUPLE,
    COMPOSITE_VECTOR_ROWS,
    COMPOSITE_VECTOR_ROWS_N,
    DISTINCT_AR_ROWS,
    INSMANY_AR_ROW_HALFN,
    INSMANY_AR_ROWS,
    INSMANY_AR_ROWS_PK_TUPLES,
    INSMANY_AR_ROWS_PKS,
    SIMPLE_FULL_ROWS,
    SIMPLE_SEVEN_ROWS_F2,
    SIMPLE_SEVEN_ROWS_F4,
    SIMPLE_SEVEN_ROWS_OK,
)


class TestTableDMLSync:
    @pytest.mark.describe("test of table insert_one and find_one, sync")
    def test_table_insert_one_find_one_sync(
        self,
        sync_empty_table_all_returns: DefaultTable,
        sync_empty_table_composite: DefaultTable,
    ) -> None:
        no_row_0a = sync_empty_table_all_returns.find_one(filter=AR_ROW_PK_0)
        assert no_row_0a is None
        ins1_res_0 = sync_empty_table_all_returns.insert_one(row=AR_ROW_0)
        row_0 = sync_empty_table_all_returns.find_one(filter=AR_ROW_PK_0)
        row_0_nofilter = sync_empty_table_all_returns.find_one(filter={})
        assert row_0 is not None
        assert row_0 == row_0_nofilter
        assert {row_0[k] == v for k, v in AR_ROW_0.items()}
        assert ins1_res_0.inserted_id == AR_ROW_PK_0
        assert _typify_tuple(ins1_res_0.inserted_id_tuple) == _typify_tuple(
            AR_ROW_PK_0_TUPLE
        )
        # overwrite:
        ins1_res_0_b = sync_empty_table_all_returns.insert_one(row=AR_ROW_0_B)
        row_0_b = sync_empty_table_all_returns.find_one(filter=AR_ROW_PK_0)
        assert row_0_b is not None
        assert {row_0_b[k] == v for k, v in AR_ROW_0_B.items()}
        assert ins1_res_0_b.inserted_id == AR_ROW_PK_0
        assert _typify_tuple(ins1_res_0_b.inserted_id_tuple) == _typify_tuple(
            AR_ROW_PK_0_TUPLE
        )
        # projection:
        projected_fields = {"p_bigint", "p_boolean"}
        row_0 = sync_empty_table_all_returns.find_one(
            filter=AR_ROW_PK_0,
            projection={pf: True for pf in projected_fields},
        )
        assert row_0 is not None
        assert {row_0[k] == AR_ROW_0[k] for k in projected_fields}
        assert row_0.keys() == projected_fields
        # delete and retry
        sync_empty_table_all_returns.delete_one(filter=AR_ROW_PK_0)
        sync_empty_table_all_returns.delete_one(filter=AR_ROW_PK_0)
        no_row_0b = sync_empty_table_all_returns.find_one(filter=AR_ROW_PK_0)
        assert no_row_0b is None

        # ANN and non-ANN sorting in find_one
        sync_empty_table_composite.insert_one(
            {
                "p_text": "pA",
                "p_int": 0,
                "p_vector": DataAPIVector([0, 1, 0]),
            }
        )
        sync_empty_table_composite.insert_many(
            [
                {
                    "p_text": "pA",
                    "p_int": i,
                    "p_vector": DataAPIVector([i, 1, 0]),
                }
                for i in range(3)
            ]
        )
        sync_empty_table_composite.insert_many(
            [
                {
                    "p_text": "pB",
                    "p_int": 100 - i,
                    "p_vector": DataAPIVector([i + 0.1, 1, 0]),
                }
                for i in range(3)
            ]
        )
        # sort by just ANN
        fo_unf_ann_row = sync_empty_table_composite.find_one(
            sort={"p_vector": DataAPIVector([-0.1, 1, 0])}
        )
        assert fo_unf_ann_row is not None
        assert (fo_unf_ann_row["p_text"], fo_unf_ann_row["p_int"]) == ("pA", 0)
        # sort by ANN, filtered
        fo_fil_ann_row = sync_empty_table_composite.find_one(
            filter={"p_text": "pB"}, sort={"p_vector": DataAPIVector([-0.1, 1, 0])}
        )
        assert fo_fil_ann_row is not None
        assert (fo_fil_ann_row["p_text"], fo_fil_ann_row["p_int"]) == ("pB", 100)
        # just regular sort
        fo_unf_srt_row = sync_empty_table_composite.find_one(
            sort={"p_int": SortMode.DESCENDING}
        )
        assert fo_unf_srt_row is not None
        assert (fo_unf_srt_row["p_text"], fo_unf_srt_row["p_int"]) == ("pB", 100)
        # regular sort, filtered
        fo_fil_srt_row = sync_empty_table_composite.find_one(
            filter={"p_text": "pA"},
            sort={"p_int": SortMode.DESCENDING},
        )
        assert fo_fil_srt_row is not None
        assert (fo_fil_srt_row["p_text"], fo_fil_srt_row["p_int"]) == ("pA", 2)

        # serdes options obeyance check
        noncustom_compo_table = sync_empty_table_composite.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            )
        )
        custom_compo_table = sync_empty_table_composite.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            )
        )
        noncustom_row = noncustom_compo_table.find_one(
            sort={"p_vector": [-0.1, 1, 0]}  # both would actually work (as arguments)
        )
        custom_row = custom_compo_table.find_one(
            sort={"p_vector": [-0.1, 1, 0]}  # both would actually work (as arguments)
        )
        assert noncustom_row is not None
        assert isinstance(noncustom_row["p_vector"], list)
        assert custom_row is not None
        assert isinstance(custom_row["p_vector"], DataAPIVector)

    @pytest.mark.describe("test of table delete_one, sync")
    def test_table_delete_one_sync(
        self,
        sync_empty_table_simple: DefaultTable,
    ) -> None:
        im_result = sync_empty_table_simple.insert_many(SIMPLE_FULL_ROWS)
        assert len(im_result.inserted_ids) == len(SIMPLE_FULL_ROWS)
        assert len(im_result.inserted_id_tuples) == len(SIMPLE_FULL_ROWS)
        assert len(sync_empty_table_simple.find({}).to_list()) == len(SIMPLE_FULL_ROWS)

        sync_empty_table_simple.delete_one({"p_text": "Z"})
        assert len(sync_empty_table_simple.find({}).to_list()) == len(SIMPLE_FULL_ROWS)

        sync_empty_table_simple.delete_one({"p_text": "A1"})
        assert (
            len(sync_empty_table_simple.find({}).to_list()) == len(SIMPLE_FULL_ROWS) - 1
        )

    @pytest.mark.describe("test of table delete_many, sync")
    def test_table_delete_many_sync(
        self,
        sync_empty_table_simple: DefaultTable,
    ) -> None:
        im_result = sync_empty_table_simple.insert_many(SIMPLE_FULL_ROWS)
        assert len(im_result.inserted_ids) == len(SIMPLE_FULL_ROWS)
        assert len(im_result.inserted_id_tuples) == len(SIMPLE_FULL_ROWS)
        assert len(sync_empty_table_simple.find({}).to_list()) == len(SIMPLE_FULL_ROWS)

        sync_empty_table_simple.delete_many({"p_text": {"$in": ["Z", "Y"]}})
        assert len(sync_empty_table_simple.find({}).to_list()) == 3

        sync_empty_table_simple.delete_many({"p_text": {"$in": ["A1", "A2"]}})
        assert len(sync_empty_table_simple.find({}).to_list()) == 1

    @pytest.mark.describe("test of table insert_many returned ids, sync")
    def test_table_insert_many_returned_ids_sync(
        self,
        sync_empty_table_all_returns: DefaultTable,
    ) -> None:
        im_result = sync_empty_table_all_returns.insert_many(INSMANY_AR_ROWS)
        assert im_result.inserted_ids == INSMANY_AR_ROWS_PKS
        assert [_typify_tuple(tpl) for tpl in im_result.inserted_id_tuples] == [
            _typify_tuple(tpl) for tpl in INSMANY_AR_ROWS_PK_TUPLES
        ]

    @pytest.mark.describe("test of table distinct, sync")
    def test_table_distinct_sync(
        self,
        sync_empty_table_all_returns: DefaultTable,
    ) -> None:
        sync_empty_table_all_returns.insert_many(DISTINCT_AR_ROWS)

        d_float = sync_empty_table_all_returns.distinct("p_float")
        exp_d_float = {0.1, 0.2, float("NaN")}
        assert set(_repaint_NaNs(d_float)) == _repaint_NaNs(exp_d_float)

        d_text = sync_empty_table_all_returns.distinct("p_text")
        exp_d_text = {"a", "b", None}
        assert set(d_text) == set(exp_d_text)

        d_timestamp = sync_empty_table_all_returns.distinct("p_timestamp")
        exp_d_timestamp = {
            DataAPITimestamp.from_string("1111-01-01T01:01:01Z"),
            DataAPITimestamp.from_string("1221-01-01T01:01:01Z"),
            None,
        }
        assert set(d_timestamp) == set(exp_d_timestamp)

        d_list_int = sync_empty_table_all_returns.distinct("p_list_int")
        exp_d_list_int = {
            itm
            for doc in DISTINCT_AR_ROWS
            for itm in cast(Iterable[int], doc.get("p_list_int", []))
        }
        assert set(d_list_int) == set(exp_d_list_int)

        d_list_int_ind = sync_empty_table_all_returns.distinct("p_list_int.1")
        exp_d_list_int_ind = {1}
        assert set(d_list_int_ind) == set(exp_d_list_int_ind)

        d_p_map_text_text = sync_empty_table_all_returns.distinct("p_map_text_text")
        exp_d_p_map_text_text = [
            {"a": "va", "b": "vb"},
            {"b": "VB"},
            {"a": "VA", "b": "VB"},
            {},
        ]
        assert len(d_p_map_text_text) == len(exp_d_p_map_text_text)
        for _exp in exp_d_p_map_text_text:
            assert any(_d == _exp for _d in d_p_map_text_text)

        d_p_map_text_text_a = sync_empty_table_all_returns.distinct("p_map_text_text.a")
        exp_d_p_map_text_text_a = {"va", "VA"}
        assert set(d_p_map_text_text_a) == set(exp_d_p_map_text_text_a)

        d_set_int = sync_empty_table_all_returns.distinct("p_set_int")
        exp_d_set_int = {
            itm
            for doc in DISTINCT_AR_ROWS
            for itm in cast(Iterable[int], doc.get("p_set_int", []))
        }
        assert set(d_set_int) == set(exp_d_set_int)

    @pytest.mark.describe("test of table distinct key-as-list, sync")
    def test_table_distinct_key_as_list_sync(
        self,
        sync_empty_table_all_returns: DefaultTable,
    ) -> None:
        sync_empty_table_all_returns.insert_many(DISTINCT_AR_ROWS)

        d_float = sync_empty_table_all_returns.distinct(["p_float"])
        exp_d_float = {0.1, 0.2, float("NaN")}
        assert set(_repaint_NaNs(d_float)) == _repaint_NaNs(exp_d_float)

        d_text = sync_empty_table_all_returns.distinct(["p_text"])
        exp_d_text = {"a", "b", None}
        assert set(d_text) == set(exp_d_text)

        d_timestamp = sync_empty_table_all_returns.distinct(["p_timestamp"])
        exp_d_timestamp = {
            DataAPITimestamp.from_string("1111-01-01T01:01:01Z"),
            DataAPITimestamp.from_string("1221-01-01T01:01:01Z"),
            None,
        }
        assert set(d_timestamp) == set(exp_d_timestamp)

        d_list_int = sync_empty_table_all_returns.distinct(["p_list_int"])
        exp_d_list_int = {
            itm
            for doc in DISTINCT_AR_ROWS
            for itm in cast(Iterable[int], doc.get("p_list_int", []))
        }
        assert set(d_list_int) == set(exp_d_list_int)

        d_list_int_ind = sync_empty_table_all_returns.distinct(["p_list_int", 1])
        exp_d_list_int_ind = {1}
        assert set(d_list_int_ind) == set(exp_d_list_int_ind)

        d_list_int_sind = sync_empty_table_all_returns.distinct(["p_list_int", "1"])
        exp_d_list_int_sind: set[int] = set()
        assert set(d_list_int_sind) == set(exp_d_list_int_sind)

        d_p_map_text_text = sync_empty_table_all_returns.distinct(["p_map_text_text"])
        exp_d_p_map_text_text = [
            {"a": "va", "b": "vb"},
            {"b": "VB"},
            {"a": "VA", "b": "VB"},
            {},
        ]
        assert len(d_p_map_text_text) == len(exp_d_p_map_text_text)
        for _exp in exp_d_p_map_text_text:
            assert any(_d == _exp for _d in d_p_map_text_text)

        d_p_map_text_text_a = sync_empty_table_all_returns.distinct(
            ["p_map_text_text", "a"]
        )
        exp_d_p_map_text_text_a = {"va", "VA"}
        assert set(d_p_map_text_text_a) == set(exp_d_p_map_text_text_a)

        d_set_int = sync_empty_table_all_returns.distinct(["p_set_int"])
        exp_d_set_int = {
            itm
            for doc in DISTINCT_AR_ROWS
            for itm in cast(Iterable[int], doc.get("p_set_int", []))
        }
        assert set(d_set_int) == set(exp_d_set_int)

    @pytest.mark.describe("test of table insert_many, sync")
    def test_table_insert_many_sync(
        self,
        sync_table_simple: DefaultTable,
    ) -> None:
        """ordered/unordered, concurrent/nonconcurrent; with good/failing rows"""

        def _pkeys() -> set[str]:
            return {
                row["p_text"]
                for row in sync_table_simple.find({}, projection={"p_text": True})
            }

        def _assert_consistency(
            p_text_values: list[str],
            ins_result: TableInsertManyResult | TableInsertManyException,
        ) -> None:
            assert _pkeys() == set(p_text_values)
            assert ins_result.inserted_id_tuples == [(pk,) for pk in p_text_values]
            assert ins_result.inserted_ids == [{"p_text": pk} for pk in p_text_values]

        def _assert_tim_exceptions(exps: Sequence[Exception], count: int) -> None:
            assert len(exps) == count
            assert all(isinstance(exp, DataAPIResponseException) for exp in exps)
            assert all(isinstance(exp.command, dict) for exp in exps)  # type: ignore[attr-defined]
            assert all(isinstance(exp.raw_response, dict) for exp in exps)  # type: ignore[attr-defined]
            assert all(exp.raw_response != exps[0].raw_response for exp in exps[1:])  # type: ignore[attr-defined]
            assert all(exp.command != exps[0].command for exp in exps[1:])  # type: ignore[attr-defined]

        # ordered, good rows
        sync_table_simple.delete_many({})
        i_result = sync_table_simple.insert_many(
            SIMPLE_SEVEN_ROWS_OK, ordered=True, chunk_size=2
        )
        _assert_consistency(["p1", "p2", "p3", "p4", "p5", "p6", "p7"], i_result)

        # ordered, failing first batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F2, ordered=True, chunk_size=2
            )
        if not os.environ.get("LEGACY_INSERTMANY_BEHAVIOUR_PRE2193"):
            _assert_consistency(["p1"], exc.value)
        else:
            _assert_consistency([], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # ordered, failing later batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F4, ordered=True, chunk_size=2
            )
        if not os.environ.get("LEGACY_INSERTMANY_BEHAVIOUR_PRE2193"):
            _assert_consistency(["p1", "p2", "p3"], exc.value)
        else:
            _assert_consistency(["p1", "p2"], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # unordered/concurrency=1, good rows
        sync_table_simple.delete_many({})
        sync_table_simple.insert_many(
            SIMPLE_SEVEN_ROWS_OK, ordered=False, chunk_size=2, concurrency=1
        )
        _assert_consistency(["p1", "p2", "p3", "p4", "p5", "p6", "p7"], i_result)

        # unordered/concurrency=1, failing first batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F2, ordered=False, chunk_size=2, concurrency=1
            )
        _assert_consistency(["p1", "p3", "p4", "p5", "p6", "p7"], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # unordered/concurrency=1, failing later batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F4, ordered=False, chunk_size=2, concurrency=1
            )
        _assert_consistency(["p1", "p2", "p3", "p5", "p6", "p7"], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # unordered/concurrency=2, good rows
        sync_table_simple.delete_many({})
        i_result = sync_table_simple.insert_many(
            SIMPLE_SEVEN_ROWS_OK, ordered=False, chunk_size=2, concurrency=2
        )
        _assert_consistency(["p1", "p2", "p3", "p4", "p5", "p6", "p7"], i_result)

        # unordered/concurrency=2, failing first batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F2, ordered=False, chunk_size=2, concurrency=2
            )
        _assert_consistency(["p1", "p3", "p4", "p5", "p6", "p7"], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # unordered/concurrency=2, failing later batch
        sync_table_simple.delete_many({})
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F4, ordered=False, chunk_size=2, concurrency=2
            )
        _assert_consistency(["p1", "p2", "p3", "p5", "p6", "p7"], exc.value)
        _assert_tim_exceptions(exc.value.exceptions, count=1)

        # more failing batches (unordered, concurrency=2)
        with pytest.raises(TableInsertManyException) as exc:
            sync_table_simple.insert_many(
                SIMPLE_SEVEN_ROWS_F4 * 2, ordered=False, chunk_size=2, concurrency=2
            )
        _assert_tim_exceptions(exc.value.exceptions, count=2)

        # with massive amount of rows:
        many_rows = [{"p_text": f"r_{i}", "p_int": i} for i in range(200)]
        exp_tuple_set = {(row["p_text"],) for row in many_rows}
        # ordered
        sync_table_simple.delete_many({})
        ins_res_o = sync_table_simple.insert_many(many_rows, ordered=True)
        assert set(ins_res_o.inserted_id_tuples) == exp_tuple_set
        assert len(sync_table_simple.find().to_list()) == len(many_rows)
        # unordered, concurrency=1
        sync_table_simple.delete_many({})
        ins_res_u_c1 = sync_table_simple.insert_many(
            many_rows, ordered=False, concurrency=1
        )
        assert set(ins_res_u_c1.inserted_id_tuples) == exp_tuple_set
        assert len(sync_table_simple.find().to_list()) == len(many_rows)
        # unordered, concurrency>1
        sync_table_simple.delete_many({})
        ins_res_u_cn = sync_table_simple.insert_many(
            many_rows, ordered=False, concurrency=10
        )
        assert set(ins_res_u_cn.inserted_id_tuples) == exp_tuple_set
        assert len(sync_table_simple.find().to_list()) == len(many_rows)

    @pytest.mark.describe("test of table update_one, sync")
    def test_table_insert_many_failures_sync(
        self,
        sync_table_simple: DefaultTable,
    ) -> None:
        # The main goal here is to keep the switch to returnDocumentResponses in check.
        N = 110
        ins_res0 = sync_table_simple.insert_many(
            [{"p_text": f"{i}"} for i in range(N)],
            concurrency=1,
        )
        assert set(ins_res0.inserted_id_tuples) == {(f"{i}",) for i in range(N)}

        ins_res1 = sync_table_simple.insert_many(
            [{"p_text": f"{N + i}"} for i in range(N)],
            concurrency=20,
        )
        assert set(ins_res1.inserted_id_tuples) == {(f"{N + i}",) for i in range(N)}

        # unordered insertion [good, bad]
        err2: TableInsertManyException | None = None
        try:
            sync_table_simple.insert_many([{"p_text": f"{2 * N}"}, {"p_text": -1}])
        except TableInsertManyException as e:
            err2 = e
        assert err2 is not None
        assert len(err2.exceptions) == 1
        assert isinstance(err2.exceptions[0], DataAPIResponseException)
        assert len(err2.exceptions[0].error_descriptors) == 1
        assert err2.inserted_id_tuples == [(f"{2 * N}",)]

        # unordered insertion [bad, bad]
        err3: TableInsertManyException | None = None
        try:
            sync_table_simple.insert_many([{"p_text": -2}, {"p_textXX": -3}])
        except TableInsertManyException as e:
            err3 = e
        assert err3 is not None
        assert len(err3.exceptions) == 1
        assert isinstance(err3.exceptions[0], DataAPIResponseException)
        assert len(err3.exceptions[0].error_descriptors) == 2
        assert err3.inserted_id_tuples == []

        # ordered insertion [good, bad, good_skipped]
        err4: TableInsertManyException | None = None
        try:
            sync_table_simple.insert_many(
                [{"p_text": "n0"}, {"p_text": -4}, {"p_text": "n1"}],
                ordered=True,
            )
        except TableInsertManyException as e:
            err4 = e
        assert err4 is not None
        assert len(err4.exceptions) == 1
        assert isinstance(err4.exceptions[0], DataAPIResponseException)
        assert len(err4.exceptions[0].error_descriptors) == 1
        if not os.environ.get("LEGACY_INSERTMANY_BEHAVIOUR_PRE2193"):
            assert err4.inserted_id_tuples == [("n0",)]
        else:
            assert err4.inserted_id_tuples == []

    @pytest.mark.describe("test of table update_one, sync")
    def test_table_update_one_sync(
        self,
        sync_table_simple: DefaultTable,
    ) -> None:
        # erroring updates
        with pytest.raises(DataAPIException):
            sync_table_simple.update_one(
                {"p_text": {"$gt": "x"}},
                update={"$set": {"p_int": -111}},
            )
        with pytest.raises(DataAPIException):
            sync_table_simple.update_one(
                {"p_text": "x"},
                update={"$set": {"p_int": -111, "p_text": "x"}},
            )
        with pytest.raises(DataAPIException):
            sync_table_simple.update_one(
                {"p_text": "x"},
                update={"$unset": {"p_text": ""}},
            )
        with pytest.raises(DataAPIException):
            sync_table_simple.update_one(
                {"p_text": "x"},
                update={"$set": {"p_fake": "x"}},
            )
        with pytest.raises(DataAPIException):
            sync_table_simple.update_one(
                {"p_int": 147},
                update={"$set": {"p_int": -111}},
            )

        # $set
        sync_table_simple.delete_many({})
        assert sync_table_simple.find_one({"p_text": "A"}) is None

        # $set, new row
        sync_table_simple.update_one(
            {"p_text": "A"},
            update={"$set": {"p_int": 10, "p_vector": DataAPIVector([1, 2, 3])}},
        )
        assert sync_table_simple.find_one({"p_text": "A"}) == {
            "p_text": "A",
            "p_int": 10,
            "p_vector": DataAPIVector([1, 2, 3]),
        }

        # $set, partial write
        sync_table_simple.update_one(
            {"p_text": "A"},
            update={"$set": {"p_int": 11}},
        )
        assert sync_table_simple.find_one({"p_text": "A"}) == {
            "p_text": "A",
            "p_int": 11,
            "p_vector": DataAPIVector([1, 2, 3]),
        }

        # $set, overwriting row
        sync_table_simple.update_one(
            {"p_text": "A"},
            update={"$set": {"p_int": 10, "p_vector": DataAPIVector([1, 2, 3])}},
        )
        assert sync_table_simple.find_one({"p_text": "A"}) == {
            "p_text": "A",
            "p_int": 10,
            "p_vector": DataAPIVector([1, 2, 3]),
        }

        # $set, partial set-to-null
        sync_table_simple.update_one(
            {"p_text": "A"},
            update={"$set": {"p_int": None}},
        )
        assert sync_table_simple.find_one({"p_text": "A"}) == {
            "p_text": "A",
            "p_int": None,
            "p_vector": DataAPIVector([1, 2, 3]),
        }

        # $set, full set-to-null (which deletes the row as it was created by update)
        sync_table_simple.update_one(
            {"p_text": "A"},
            update={"$set": {"p_int": None, "p_vector": None}},
        )
        assert sync_table_simple.find_one({"p_text": "A"}) is None

        # $unset
        sync_table_simple.delete_many({})
        sync_table_simple.insert_one(
            {"p_text": "B", "p_int": 123, "p_vector": DataAPIVector([9, 8, 7])}
        )
        assert sync_table_simple.find_one({"p_text": "B"}) == {
            "p_text": "B",
            "p_int": 123,
            "p_vector": DataAPIVector([9, 8, 7]),
        }

        # $unset, on existing(partial)
        sync_table_simple.update_one(
            {"p_text": "B"},
            update={"$unset": {"p_int": ""}},
        )
        assert sync_table_simple.find_one({"p_text": "B"}) == {
            "p_text": "B",
            "p_int": None,
            "p_vector": DataAPIVector([9, 8, 7]),
        }

        # $unset, on existing(partial) 2: the other column
        sync_table_simple.update_one(
            {"p_text": "B"},
            update={"$unset": {"p_vector": ""}},
        )
        assert sync_table_simple.find_one({"p_text": "B"}) == {
            "p_text": "B",
            "p_int": None,
            "p_vector": None,
        }

        # $unset, on existing(full) (which DOES NOT DELETE the row)
        sync_table_simple.update_one(
            {"p_text": "B"},
            update={"$unset": {"p_int": "", "p_vector": ""}},
        )
        assert sync_table_simple.find_one({"p_text": "B"}) == {
            "p_text": "B",
            "p_int": None,
            "p_vector": None,
        }

        # $unset, on nonexisting
        sync_table_simple.update_one(
            {"p_text": "Z"},
            update={"$unset": {"p_int": ""}},
        )
        assert sync_table_simple.find_one({"p_text": "Z"}) is None

    @pytest.mark.describe("test of include_sort_vector with serdes options, sync")
    def test_table_include_sort_vector_serdes_options_sync(
        self,
        sync_table_simple: DefaultTable,
    ) -> None:
        col_v0 = sync_table_simple.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            ),
        )
        col_v1 = sync_table_simple.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            ),
        )
        cur0_v0_inpf = col_v0.find(sort={"p_vector": [1, 2, 3]})
        cur0_v1_inpf = col_v1.find(sort={"p_vector": [1, 2, 3]})

        assert cur0_v0_inpf.get_sort_vector() is None
        assert cur0_v1_inpf.get_sort_vector() is None

        cur1_v0_inpf = col_v0.find(
            sort={"p_vector": [1, 2, 3]}, include_sort_vector=True
        )
        cur1_v1_inpf = col_v1.find(
            sort={"p_vector": [1, 2, 3]}, include_sort_vector=True
        )

        assert cur1_v0_inpf.get_sort_vector() == [1, 2, 3]
        assert cur1_v1_inpf.get_sort_vector() == DataAPIVector([1, 2, 3])

        cur0_v0_inpv = col_v0.find(sort={"p_vector": DataAPIVector([1, 2, 3])})
        cur0_v1_inpv = col_v1.find(sort={"p_vector": DataAPIVector([1, 2, 3])})

        assert cur0_v0_inpv.get_sort_vector() is None
        assert cur0_v1_inpv.get_sort_vector() is None

        cur1_v0_inpv = col_v0.find(
            sort={"p_vector": DataAPIVector([1, 2, 3])}, include_sort_vector=True
        )
        cur1_v1_inpv = col_v1.find(
            sort={"p_vector": DataAPIVector([1, 2, 3])}, include_sort_vector=True
        )

        assert cur1_v0_inpv.get_sort_vector() == [1, 2, 3]
        assert cur1_v1_inpv.get_sort_vector() == DataAPIVector([1, 2, 3])

    @pytest.mark.describe("test of table find, sync")
    def test_table_find_sync(
        self,
        sync_empty_table_composite: DefaultTable,
        sync_empty_table_all_returns: DefaultTable,
    ) -> None:
        sync_empty_table_composite.insert_many(
            [
                {
                    "p_text": "pA",
                    "p_int": i,
                    "p_boolean": i % 2 == 0,
                    "p_vector": DataAPIVector([i, 5, 6]),
                }
                for i in range(120)
            ]
        )
        sync_empty_table_composite.insert_many(
            [
                {
                    "p_text": "pB",
                    "p_int": i,
                    "p_boolean": i % 2 == 0,
                    "p_vector": DataAPIVector([i, 6, 5]),
                }
                for i in range(120)
            ]
        )

        # partition filter
        rows_a = sync_empty_table_composite.find({"p_text": "pA"}).to_list()
        assert len(rows_a) == 120
        assert all(row["p_text"] == "pA" for row in rows_a)
        # no filters
        rows_all = sync_empty_table_composite.find({}).to_list()
        assert len(rows_all) == 240
        # TODO: enable for Astra once #2089 gets to Astra
        if not IS_ASTRA_DB:
            # a logically-combined condition on the partition key should fail from DB:
            with pytest.raises(DataAPIResponseException):
                sync_empty_table_composite.find(
                    {"$or": [{"p_text": "pA"}, {"p_text": "pB"}]}
                ).to_list()
        # non-pk-column filter, alone
        rows_even_allps = sync_empty_table_composite.find({"p_boolean": True}).to_list()
        assert len(rows_even_allps) == 2 * sum(1 - i % 2 for i in range(120))
        assert all(row["p_boolean"] for row in rows_even_allps)
        # non-pk-column + partition key filter
        rows_even_a = sync_empty_table_composite.find(
            {"p_text": "pA", "p_boolean": True}
        ).to_list()
        assert len(rows_even_a) == sum(1 - i % 2 for i in range(120))
        assert all(row["p_text"] == "pA" for row in rows_even_a)
        assert all(row["p_boolean"] for row in rows_even_a)

        # projection
        projected_fields = {"p_int", "p_vector"}
        rows_proj_a = sync_empty_table_composite.find(
            {"p_text": "pA"},
            projection={f: True for f in projected_fields},
        ).to_list()
        assert all(row.keys() == projected_fields for row in rows_proj_a)

        # (nonvector) sorting
        sync_empty_table_all_returns.insert_many(INSMANY_AR_ROWS)
        # sorting as per clustering column
        srows_in_part = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 100},
            sort={"p_int": SortMode.DESCENDING},
            limit=INSMANY_AR_ROW_HALFN + 1,
        ).to_list()
        assert len(srows_in_part) == INSMANY_AR_ROW_HALFN
        srows_in_part_pints = [row["p_int"] for row in srows_in_part]
        assert sorted(srows_in_part_pints) == srows_in_part_pints[::-1]
        # sorting by any regular column
        srows_anycol = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 100},
            sort={"p_float": SortMode.DESCENDING},
            limit=INSMANY_AR_ROW_HALFN + 1,
        ).to_list()
        # sorted finds in this case return at most one page and that's it:
        assert len(srows_anycol) == 20
        srows_anycol_pints = [row["p_int"] for row in srows_anycol]
        assert sorted(srows_anycol_pints) == srows_anycol_pints[::-1]

        # use of limit+skip: two ways of getting the first 30 items
        ls_rows_A0 = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 200},
            sort={"p_int": SortMode.DESCENDING},
            skip=0,
            limit=20,
        ).to_list()
        ls_rows_A1 = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 200},
            sort={"p_int": SortMode.DESCENDING},
            skip=20,
            limit=10,
        ).to_list()
        ls_rows_B0 = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 200},
            sort={"p_int": SortMode.DESCENDING},
            skip=0,
            limit=10,
        ).to_list()
        ls_rows_B1 = sync_empty_table_all_returns.find(
            filter={"p_ascii": "A", "p_bigint": 200},
            sort={"p_int": SortMode.DESCENDING},
            skip=10,
            limit=20,
        ).to_list()
        ls_rows_A = ls_rows_A0 + ls_rows_A1
        ls_rows_B = ls_rows_B0 + ls_rows_B1
        assert len(ls_rows_A) == 30
        assert ls_rows_A == ls_rows_B

        # find with ANN
        sync_empty_table_composite.delete_many({})
        sync_empty_table_composite.insert_many(COMPOSITE_VECTOR_ROWS)
        # in a partition
        vrows_in_part = sync_empty_table_composite.find(
            filter={"p_text": "A"},
            sort={"p_vector": DataAPIVector([COMPOSITE_VECTOR_ROWS_N, 0, 0])},
            limit=COMPOSITE_VECTOR_ROWS_N + 2,
        ).to_list()
        assert len(vrows_in_part) == COMPOSITE_VECTOR_ROWS_N
        ints = [row["p_int"] for row in vrows_in_part]
        assert sorted(ints, reverse=True) == ints
        # across all partitions
        vrows_in_part = sync_empty_table_composite.find(
            sort={"p_vector": DataAPIVector([COMPOSITE_VECTOR_ROWS_N, 0, 0])},
            limit=2 * COMPOSITE_VECTOR_ROWS_N + 2,
        ).to_list()
        assert len(vrows_in_part) == 2 * COMPOSITE_VECTOR_ROWS_N
        ints = [row["p_int"] for row in vrows_in_part]
        assert sorted(ints, reverse=True) == ints
        # filtering on a non-pk column
        vrows_in_part = sync_empty_table_composite.find(
            filter={"p_boolean": True},
            sort={"p_vector": DataAPIVector([COMPOSITE_VECTOR_ROWS_N, 0, 0])},
            limit=2 * COMPOSITE_VECTOR_ROWS_N + 2,
        ).to_list()
        assert len(vrows_in_part) == 2 * sum(
            1 - i % 2 for i in range(COMPOSITE_VECTOR_ROWS_N)
        )
        ints = [row["p_int"] for row in vrows_in_part]
        assert all(i % 2 == 0 for i in ints)
        assert sorted(ints, reverse=True) == ints

        # mapper and to_list at once
        def _mapping(dct: dict[str, Any]) -> dict[str, str]:
            return {k: str(v) for k, v in dct.items()}

        mapped_cursor = sync_empty_table_composite.find(
            sort={"p_vector": DataAPIVector([1, 0, 0])},
            limit=COMPOSITE_VECTOR_ROWS_N,
        ).map(_mapping)
        unmapped_list = sync_empty_table_composite.find(
            sort={"p_vector": DataAPIVector([1, 0, 0])},
            limit=COMPOSITE_VECTOR_ROWS_N,
        ).to_list()
        postmapped_list = [_mapping(row) for row in unmapped_list]
        premapped_list = [mrow for mrow in mapped_cursor]
        assert postmapped_list == premapped_list

    # TODO: enable for Astra once #2089 gets to Astra
    @pytest.mark.skipif(IS_ASTRA_DB, reason="Feature not on Astra DB yet")
    @pytest.mark.describe("test of table logical filtering, sync")
    def test_table_logicalfiltering_sync(
        self,
        sync_empty_table_logicalfiltering: DefaultTable,
    ) -> None:
        table = sync_empty_table_logicalfiltering
        table.insert_one(
            {
                "id": "base",
                "p_boolean": False,
                "p_int_1": 10,
                "p_int_2": 100,
                "p_text": "Copper",
            }
        )

        # match {and: {or: (f,t), or: (t,f)}}
        row_0 = table.find_one(
            {
                "$and": [
                    {"$or": [{"p_boolean": True}, {"p_int_1": {"$lt": 15}}]},
                    {
                        "$or": [
                            {"p_text": {"$in": ["Iron", "Copper"]}},
                            {"p_int_2": {"$gte": 150}},
                        ]
                    },
                ]
            }
        )
        assert row_0 is not None
        assert row_0["id"] == "base"

        # no {and: {or: (f,f), or: (t,f)}}
        row_0_n = table.find_one(
            {
                "$and": [
                    {"$or": [{"p_boolean": True}, {"p_int_1": {"$lt": 5}}]},
                    {
                        "$or": [
                            {"p_text": {"$in": ["Iron", "Copper"]}},
                            {"p_int_2": {"$gte": 50}},
                        ]
                    },
                ]
            }
        )
        assert row_0_n is None

        table.insert_many(
            [
                {
                    "id": f"i2_{i:02}",
                    "p_int_2": i,
                }
                for i in range(20)
            ]
        )

        # range find
        rows_range = table.find(
            {
                "$and": [
                    {"p_int_2": {"$gte": 10}},
                    {"p_int_2": {"$lt": 15}},
                ]
            }
        ).to_list()
        assert {row["p_int_2"] for row in rows_range} == {10, 11, 12, 13, 14}

    @pytest.mark.describe("test of table command, sync")
    def test_table_command_sync(
        self,
        sync_empty_table_simple: DefaultTable,
    ) -> None:
        ins_response = sync_empty_table_simple.command(
            body={
                "insertOne": {
                    "document": {
                        "p_text": "t_command",
                        "p_int": 101,
                        "p_vector": [0.1, -0.2, 0.3],
                    }
                }
            }
        )
        assert ins_response == {
            "status": {
                "primaryKeySchema": {"p_text": {"type": "text"}},
                "insertedIds": [["t_command"]],
            }
        }



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum, EnumMeta
from typing import TypeVar

T = TypeVar("T", bound="StrEnum")


class StrEnumMeta(EnumMeta):
    def _name_lookup(cls, value: str) -> str | None:
        """Return a proper key in the enum if some matching logic works, or None."""
        mmap = {k: v.value for k, v in cls._member_map_.items()}
        # try exact key match
        if value in mmap:
            return value
        # try case-insensitive key match
        u_value = value.upper()
        u_mmap = {k.upper(): k for k in mmap.keys()}
        if u_value in u_mmap:
            return u_mmap[u_value]
        # try case-insensitive *value* match
        v_mmap = {v.upper(): k for k, v in mmap.items()}
        if u_value in v_mmap:
            return v_mmap[u_value]
        return None

    def __contains__(cls, value: object) -> bool:
        """Return True if the provided string belongs to the enum."""
        if isinstance(value, str):
            return cls._name_lookup(value) is not None
        return False


class StrEnum(Enum, metaclass=StrEnumMeta):
    @classmethod
    def coerce(cls: type[T], value: str | T) -> T:
        """
        Accepts either a string or an instance of the Enum itself.
        If a string is passed, it converts it to the corresponding
        Enum value (case-insensitive).
        If an Enum instance is passed, it returns it as-is.
        Raises ValueError if the string does not match any enum member.
        """

        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            norm_value = cls._name_lookup(value)
            if norm_value is not None:
                return cls[norm_value]
            # no matches
            raise ValueError(
                f"Invalid value '{value}' for {cls.__name__}. "
                f"Allowed values are: {[e.value for e in cls]}"
            )
        raise ValueError(
            f"Invalid value '{value}' for {cls.__name__}. "
            f"Allowed values are: {[e.value for e in cls]}"
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.data.utils.distinct_extractors import (
    _create_document_key_extractor,
    _reduce_distinct_key_to_safe,
    _reduce_distinct_key_to_shallow_safe,
)

DOCUMENT_1 = {
    "l": ["L0", "L1", {"name": "L2"}],
    "d": {"0": "D0", "1": "D1", "2": {"name": "D2"}},
    "ll": [
        ["LL00", "LL01"],
        ["LL10", {"name": "LL11"}],
    ],
    "ld": [
        {"0": "LD00", "1": "LD01"},
        {"0": "LD10", "1": {"name": "LD11"}},
    ],
    "dl": {
        "0": ["DL090", "DL01"],
        "1": ["DL10", {"name": "DL11"}],
    },
    "dd": {
        "0": {"0": "DL00", "1": "DL01"},
        "1": {"0": "DL10", "1": {"name": "DL11"}},
    },
    "non_index": {"00": "NI00", "01": "NI01"},
}
DOCUMENT_2A = {"x": [{"y": 10}, {"y": 11}, {"y": 12}]}
DOCUMENT_2B = {"x": [{"y": 100}]}
DOCUMENT_2C = {"x": ["X0", "X1"]}
DOCUMENT_3 = {"a": [1, 2, ["a", "b", "c", "d", "e", "f"]]}
DOCUMENT_ESC_1 = {
    "a.b": {"..": "the-dots"},
    "c&d": [
        {"&&": "the-ampersands"},
        {"..": "amp-then-dots"},
    ],
    ".": {"0": "dot-zero"},
    "&": ["amp-zero"],
}


def _assert_extracts(
    document: dict[str, Any], key: str | list[str | int], expected: list[Any]
) -> None:
    _extractor = _create_document_key_extractor(key)
    _extracted = list(_extractor(document))
    assert len(_extracted) == len(expected)
    for item in expected:
        assert item in _extracted


class TestDocumentExtractors:
    @pytest.mark.describe("test of key reduction to safe, from str")
    def test_reduce_key_to_safe_str(self) -> None:
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_safe("")
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_safe("0.a")
        assert _reduce_distinct_key_to_safe("a.b.0.d") == "a.b"
        assert _reduce_distinct_key_to_safe("a.b") == "a.b"
        assert _reduce_distinct_key_to_safe("a") == "a"

    @pytest.mark.describe("test of key reduction to safe, from list")
    def test_reduce_key_to_safe_list(self) -> None:
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_safe([])
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_safe([0, "a"])
        assert _reduce_distinct_key_to_safe(["a", "b", 0, "d"]) == "a.b"
        assert _reduce_distinct_key_to_safe(["a", "b"]) == "a.b"
        assert _reduce_distinct_key_to_safe(["a"]) == "a"

    @pytest.mark.describe("test of key reduction to shallow safe, from str")
    def test_reduce_key_to_shallow_safe_str(self) -> None:
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_shallow_safe("")
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_shallow_safe("0.a")
        assert _reduce_distinct_key_to_shallow_safe("a.b.0.d") == "a"
        assert _reduce_distinct_key_to_shallow_safe("a.b") == "a"
        assert _reduce_distinct_key_to_shallow_safe("a") == "a"

    @pytest.mark.describe("test of key reduction to shallow safe, from list")
    def test_reduce_key_to_shallow_safe_list(self) -> None:
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_shallow_safe([])
        with pytest.raises(ValueError):
            _reduce_distinct_key_to_shallow_safe([0, "a"])
        assert _reduce_distinct_key_to_shallow_safe(["a", "b", 0, "d"]) == "a"
        assert _reduce_distinct_key_to_shallow_safe(["a", "b"]) == "a"
        assert _reduce_distinct_key_to_shallow_safe(["a"]) == "a"

    @pytest.mark.describe("test of plain fieldname document extractor, from str")
    def test_plain_fieldname_document_extractor_str(self) -> None:
        f_extractor = _create_document_key_extractor("f")

        assert list(f_extractor({})) == []
        assert list(f_extractor({"e": "E", "f": "F", "g": "G"})) == ["F"]
        assert list(f_extractor({"f": {"name": "F"}})) == [{"name": "F"}]
        assert list(f_extractor({"f": None})) == [None]

    @pytest.mark.describe("test of dotted fieldname document extractor, from str")
    def test_dotted_fieldname_document_extractor_str(self) -> None:
        # various patterns mixing dict and list indexed access, with string input

        _assert_extracts(DOCUMENT_1, "l", ["L0", "L1", {"name": "L2"}])
        _assert_extracts(DOCUMENT_1, "l.1", ["L1"])
        _assert_extracts(DOCUMENT_1, "l.2", [{"name": "L2"}])
        _assert_extracts(DOCUMENT_1, "d", [{"0": "D0", "1": "D1", "2": {"name": "D2"}}])
        _assert_extracts(DOCUMENT_1, "d.1", ["D1"])
        _assert_extracts(DOCUMENT_1, "d.2", [{"name": "D2"}])
        _assert_extracts(
            DOCUMENT_1, "ll", [["LL00", "LL01"], ["LL10", {"name": "LL11"}]]
        )
        _assert_extracts(DOCUMENT_1, "ll.1", ["LL10", {"name": "LL11"}])
        _assert_extracts(DOCUMENT_1, "ll.1.1", [{"name": "LL11"}])
        _assert_extracts(
            DOCUMENT_1,
            "ld",
            [{"0": "LD00", "1": "LD01"}, {"0": "LD10", "1": {"name": "LD11"}}],
        )
        _assert_extracts(DOCUMENT_1, "ld.1", [{"0": "LD10", "1": {"name": "LD11"}}])
        _assert_extracts(DOCUMENT_1, "ld.1.1", [{"name": "LD11"}])
        _assert_extracts(
            DOCUMENT_1,
            "dl",
            [{"0": ["DL090", "DL01"], "1": ["DL10", {"name": "DL11"}]}],
        )
        _assert_extracts(DOCUMENT_1, "dl.1", ["DL10", {"name": "DL11"}])
        _assert_extracts(DOCUMENT_1, "dl.1.1", [{"name": "DL11"}])
        _assert_extracts(
            DOCUMENT_1,
            "dd",
            [
                {
                    "0": {"0": "DL00", "1": "DL01"},
                    "1": {"0": "DL10", "1": {"name": "DL11"}},
                }
            ],
        )
        _assert_extracts(DOCUMENT_1, "dd.1", [{"0": "DL10", "1": {"name": "DL11"}}])
        _assert_extracts(DOCUMENT_1, "dd.1.1", [{"name": "DL11"}])

        # test about omitting list indices
        _assert_extracts(DOCUMENT_2A, "x.y", [10, 11, 12])
        _assert_extracts(DOCUMENT_2A, "x.0.y", [10])
        _assert_extracts(DOCUMENT_2A, "x.1.y", [11])

        _assert_extracts(DOCUMENT_2B, "x.y", [100])
        _assert_extracts(DOCUMENT_2B, "x.0.y", [100])
        _assert_extracts(DOCUMENT_2B, "x.1.y", [])

        _assert_extracts(DOCUMENT_2C, "x", ["X0", "X1"])
        _assert_extracts(DOCUMENT_2C, "x.0", ["X0"])
        _assert_extracts(DOCUMENT_2C, "x.1", ["X1"])

        # nonexistent index should block list auto-unrolling
        _assert_extracts(DOCUMENT_3, "a.5", [])

    @pytest.mark.describe("test of dotted fieldname document extractor, from list")
    def test_dotted_fieldname_document_extractor_list(self) -> None:
        # various patterns mixing dict and list indexed access, with list input

        _assert_extracts(DOCUMENT_1, ["l"], ["L0", "L1", {"name": "L2"}])
        _assert_extracts(DOCUMENT_1, ["l", 1], ["L1"])
        _assert_extracts(DOCUMENT_1, ["l", "1"], [])
        _assert_extracts(DOCUMENT_1, ["l", 2], [{"name": "L2"}])
        _assert_extracts(DOCUMENT_1, ["l", "2"], [])
        _assert_extracts(
            DOCUMENT_1, ["d"], [{"0": "D0", "1": "D1", "2": {"name": "D2"}}]
        )
        _assert_extracts(DOCUMENT_1, ["d", "1"], ["D1"])
        _assert_extracts(DOCUMENT_1, ["d", 1], [])
        _assert_extracts(DOCUMENT_1, ["d", "2"], [{"name": "D2"}])
        _assert_extracts(DOCUMENT_1, ["d", 2], [])
        _assert_extracts(
            DOCUMENT_1, ["ll"], [["LL00", "LL01"], ["LL10", {"name": "LL11"}]]
        )
        _assert_extracts(DOCUMENT_1, ["ll", 1], ["LL10", {"name": "LL11"}])
        _assert_extracts(DOCUMENT_1, ["ll", 1, 1], [{"name": "LL11"}])
        _assert_extracts(
            DOCUMENT_1,
            ["ld"],
            [{"0": "LD00", "1": "LD01"}, {"0": "LD10", "1": {"name": "LD11"}}],
        )
        _assert_extracts(DOCUMENT_1, ["ld", 1], [{"0": "LD10", "1": {"name": "LD11"}}])
        _assert_extracts(DOCUMENT_1, ["ld", "1"], ["LD01", {"name": "LD11"}])
        _assert_extracts(DOCUMENT_1, ["ld", 1, "1"], [{"name": "LD11"}])
        _assert_extracts(DOCUMENT_1, ["ld", 1, 1], [])
        _assert_extracts(DOCUMENT_1, ["ld", "1", 1], [])
        _assert_extracts(DOCUMENT_1, ["ld", "1", "1"], [])
        _assert_extracts(
            DOCUMENT_1,
            ["dl"],
            [{"0": ["DL090", "DL01"], "1": ["DL10", {"name": "DL11"}]}],
        )
        _assert_extracts(DOCUMENT_1, ["dl", "1"], ["DL10", {"name": "DL11"}])
        _assert_extracts(DOCUMENT_1, ["dl", 1], [])
        _assert_extracts(DOCUMENT_1, ["dl", "1", 1], [{"name": "DL11"}])
        _assert_extracts(DOCUMENT_1, ["dl", "1", "1"], [])
        _assert_extracts(DOCUMENT_1, ["dl", 1, 1], [])
        _assert_extracts(DOCUMENT_1, ["dl", 1, "1"], [])
        _assert_extracts(
            DOCUMENT_1,
            ["dd"],
            [
                {
                    "0": {"0": "DL00", "1": "DL01"},
                    "1": {"0": "DL10", "1": {"name": "DL11"}},
                }
            ],
        )
        _assert_extracts(
            DOCUMENT_1, ["dd", "1"], [{"0": "DL10", "1": {"name": "DL11"}}]
        )
        _assert_extracts(DOCUMENT_1, ["dd", 1], [])
        _assert_extracts(DOCUMENT_1, ["dd", "1", "1"], [{"name": "DL11"}])
        _assert_extracts(DOCUMENT_1, ["dd", "1", 1], [])
        _assert_extracts(DOCUMENT_1, ["dd", 1, "1"], [])
        _assert_extracts(DOCUMENT_1, ["dd", 1, 1], [])

        # test about omitting list indices
        _assert_extracts(DOCUMENT_2A, ["x", "y"], [10, 11, 12])
        _assert_extracts(DOCUMENT_2A, ["x", 0, "y"], [10])
        _assert_extracts(DOCUMENT_2A, ["x", 1, "y"], [11])

        _assert_extracts(DOCUMENT_2B, ["x", "y"], [100])
        _assert_extracts(DOCUMENT_2B, ["x", 0, "y"], [100])
        _assert_extracts(DOCUMENT_2B, ["x", 1, "y"], [])

        _assert_extracts(DOCUMENT_2C, ["x"], ["X0", "X1"])
        _assert_extracts(DOCUMENT_2C, ["x", 0], ["X0"])
        _assert_extracts(DOCUMENT_2C, ["x", 1], ["X1"])

        # nonexistent index should block list auto-unrolling
        _assert_extracts(DOCUMENT_3, ["a", 5], [])

    @pytest.mark.describe("test of escapee fieldname document extractor, from str")
    def test_escapee_fieldname_document_extractor_str(self) -> None:
        # field names that ought to be escaped, str (escaped) specification
        _assert_extracts(DOCUMENT_ESC_1, "a&.b.&.&.", ["the-dots"])
        _assert_extracts(DOCUMENT_ESC_1, "c&&d.&&&&", ["the-ampersands"])
        _assert_extracts(DOCUMENT_ESC_1, "c&&d.0.&&&&", ["the-ampersands"])
        _assert_extracts(DOCUMENT_ESC_1, "c&&d.&.&.", ["amp-then-dots"])
        _assert_extracts(DOCUMENT_ESC_1, "c&&d.1.&.&.", ["amp-then-dots"])
        _assert_extracts(DOCUMENT_ESC_1, "c&&d.0.&.&.", [])
        _assert_extracts(DOCUMENT_ESC_1, "&..0", ["dot-zero"])
        _assert_extracts(DOCUMENT_ESC_1, "&&.0", ["amp-zero"])

    @pytest.mark.describe("test of escapee fieldname document extractor, from list")
    def test_escapee_fieldname_document_extractor_list(self) -> None:
        # field names that ought to be escaped, literals specification
        _assert_extracts(DOCUMENT_ESC_1, ["a.b", ".."], ["the-dots"])
        _assert_extracts(DOCUMENT_ESC_1, ["c&d", "&&"], ["the-ampersands"])
        _assert_extracts(DOCUMENT_ESC_1, ["c&d", 0, "&&"], ["the-ampersands"])
        _assert_extracts(DOCUMENT_ESC_1, ["c&d", ".."], ["amp-then-dots"])
        _assert_extracts(DOCUMENT_ESC_1, ["c&d", 1, ".."], ["amp-then-dots"])
        _assert_extracts(DOCUMENT_ESC_1, ["c&d", 0, ".."], [])
        _assert_extracts(DOCUMENT_ESC_1, [".", "0"], ["dot-zero"])
        _assert_extracts(DOCUMENT_ESC_1, [".", 0], [])
        _assert_extracts(DOCUMENT_ESC_1, ["&", "0"], [])
        _assert_extracts(DOCUMENT_ESC_1, ["&", 0], ["amp-zero"])

    @pytest.mark.describe("test of failure modes in creating document extractors")
    def test_failure_document_extractors(self) -> None:
        _create_document_key_extractor([0, 1, "two"])
        _create_document_key_extractor("0.1.two")

        with pytest.raises(ValueError):
            _create_document_key_extractor(".xyz")
        with pytest.raises(ValueError):
            _create_document_key_extractor("")
        with pytest.raises(ValueError):
            _create_document_key_extractor("a.b..d")
        with pytest.raises(ValueError):
            _create_document_key_extractor(["", "xyz"])
        with pytest.raises(ValueError):
            _create_document_key_extractor([])
        with pytest.raises(ValueError):
            _create_document_key_extractor(["a", "b", "", "d"])



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import ipaddress
from decimal import Decimal
from typing import Any

from astrapy.data_types import (
    DataAPIDate,
    DataAPIDictUDT,
    DataAPIDuration,
    DataAPIMap,
    DataAPISet,
    DataAPITime,
    DataAPITimestamp,
    DataAPIVector,
)
from astrapy.ids import UUID
from astrapy.info import (
    AlterTypeAddFields,
    AlterTypeRenameFields,
    CreateTypeDefinition,
)

AR_ROW_PK_0 = {
    "p_ascii": "abc",
    "p_bigint": 10000,
    "p_int": 987,
    "p_boolean": False,
}
AR_ROW_PK_0_TUPLE = ("abc", 10000, 987, False)
AR_ROW_0 = {
    "p_text": "Ålesund",
    **AR_ROW_PK_0,
}
AR_ROW_0_B = {
    "p_text": "Overwritten™",
    **AR_ROW_PK_0,
}

DISTINCT_AR_ROWS = [
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 1,
        "p_boolean": True,
        #
        "p_float": 0.1,
        "p_text": "a",
        "p_timestamp": DataAPITimestamp.from_string("1111-01-01T01:01:01Z"),
        "p_list_int": [1, 1, 2],
        "p_map_text_text": {"a": "va", "b": "vb"},
        "p_set_int": {100, 200},
    },
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 2,
        "p_boolean": True,
        #
        "p_float": 0.1,
        "p_text": "a",
        "p_timestamp": DataAPITimestamp.from_string("1111-01-01T01:01:01Z"),
        "p_list_int": [2, 1],
        "p_map_text_text": {"a": "va", "b": "vb"},
        "p_set_int": {200},
    },
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 3,
        "p_boolean": True,
        #
        "p_float": float("NaN"),
        "p_text": "a",
        "p_list_int": [],
        "p_map_text_text": {"b": "VB"},
        "p_set_int": set([]),
    },
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 4,
        "p_boolean": True,
        #
        "p_float": float("NaN"),
    },
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 5,
        "p_boolean": True,
        #
        "p_float": 0.2,
        "p_text": "b",
        "p_timestamp": DataAPITimestamp.from_string("1221-01-01T01:01:01Z"),
        "p_list_int": [3, 1],
        "p_map_text_text": {"a": "VA", "b": "VB"},
        "p_set_int": {200, 300},
    },
]
DISTINCT_EMPTYMAP_AR_ROWS = [
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 6,
        "p_boolean": True,
        #
        "p_list_int": [],
        "p_map_text_text": {},
        "p_set_int": set(),
    },
    {
        "p_ascii": "A",
        "p_bigint": 1,
        "p_int": 7,
        "p_boolean": True,
        #
        "p_list_int": [],
        "p_map_text_text": DataAPIMap([]),
        "p_set_int": set(),
    },
]

INSMANY_AR_ROW_HALFN = 250
INSMANY_AR_ROWS = [
    {
        "p_ascii": "A",
        "p_bigint": (PART_I + 1) * 100,
        "p_int": i,
        "p_boolean": i % 2 == 0,
        #
        "p_vector": None,
        "p_float": 0.2 + i,
        "p_text": f"b{i}",
        "p_timestamp": DataAPITimestamp.from_string(f"1{i:03}-01-01T01:01:01Z"),
        "p_list_int": [3 + i, 1, i],
        "p_map_text_text": {"a": f"V{i}A", "b": f"V{i}B"},
        "p_set_int": {200 + i, 300, i},
    }
    for PART_I in range(2)
    for i in range(INSMANY_AR_ROW_HALFN)
]
INSMANY_AR_ROWS_PKS = [
    {
        "p_ascii": ar_row["p_ascii"],
        "p_bigint": ar_row["p_bigint"],
        "p_int": ar_row["p_int"],
        "p_boolean": ar_row["p_boolean"],
    }
    for ar_row in INSMANY_AR_ROWS
]
INSMANY_AR_ROWS_PK_TUPLES = [
    (ar_row["p_ascii"], ar_row["p_bigint"], ar_row["p_int"], ar_row["p_boolean"])
    for ar_row in INSMANY_AR_ROWS
]

SIMPLE_FULL_ROWS = [
    {"p_text": "A1", "p_int": 1, "p_vector": [1.1, 1.1, 1.1]},
    {"p_text": "A2", "p_int": 2, "p_vector": [2.2, 2.2, 2.2]},
    {"p_text": "A3", "p_int": 3, "p_vector": [3.3, 3.3, 3.3]},
]

SIMPLE_SEVEN_ROWS_OK: list[dict[str, Any]] = [
    {"p_text": "p1", "p_int": 1},
    {"p_text": "p2", "p_int": 2},
    {"p_text": "p3", "p_int": 3},
    {"p_text": "p4", "p_int": 4},
    {"p_text": "p5", "p_int": 5},
    {"p_text": "p6", "p_int": 6},
    {"p_text": "p7", "p_int": 7},
]

SIMPLE_SEVEN_ROWS_F2: list[dict[str, Any]] = [
    {"p_text": "p1", "p_int": 1},
    {"p_text": "p2", "p_int": "boo"},
    {"p_text": "p3", "p_int": 3},
    {"p_text": "p4", "p_int": 4},
    {"p_text": "p5", "p_int": 5},
    {"p_text": "p6", "p_int": 6},
    {"p_text": "p7", "p_int": 7},
]

SIMPLE_SEVEN_ROWS_F4: list[dict[str, Any]] = [
    {"p_text": "p1", "p_int": 1},
    {"p_text": "p2", "p_int": 2},
    {"p_text": "p3", "p_int": 3},
    {"p_text": "p4", "p_int": "boo"},
    {"p_text": "p5", "p_int": 5},
    {"p_text": "p6", "p_int": 6},
    {"p_text": "p7", "p_int": 7},
]

FULL_AR_ROW_CUSTOMTYPED = {
    "p_ascii": "A",
    "p_bigint": 1230000,
    "p_blob": b"xyz",
    "p_boolean": True,
    "p_date": DataAPIDate.from_string("9876-01-12"),
    "p_decimal": Decimal("123.456"),
    "p_double": 123.5555,
    "p_duration": DataAPIDuration.from_c_string("4d3h55ns"),
    "p_float": 1.345,
    "p_inet": ipaddress.ip_address("127.0.0.1"),
    "p_int": 1012,
    "p_smallint": 12,
    "p_text": "Àäxxyxy",
    "p_text_nulled": None,
    "p_text_omitted": None,
    "p_time": DataAPITime.from_string("12:34:56.789"),
    "p_timestamp": DataAPITimestamp.from_string("1998-07-14T12:34:56.789Z"),
    "p_tinyint": 7,
    "p_varint": 10293,
    "p_uuid": UUID("01932c57-b8b7-8310-a84e-6d3fada3c525"),
    "p_vector": DataAPIVector([-0.1, 0.2, -0.3]),
    "p_list_int": [10, 11, -12],
    "p_map_text_text": DataAPIMap({"x": "YYY", "p": "QQQ"}),
    "p_set_int": DataAPISet({-43, 111, 109, 0}),
    "p_double_minf": float("-Infinity"),
    "p_double_pinf": float("Infinity"),
    "p_float_nan": float("NaN"),
}
FULL_AR_ROW_NONCUSTOMTYPED = {
    "p_ascii": "A",
    "p_bigint": 1230000,
    "p_blob": b"xyz",
    "p_boolean": True,
    "p_date": datetime.date(2013, 7, 24),
    "p_decimal": Decimal("123.456"),
    "p_double": 123.5555,
    "p_duration": datetime.timedelta(hours=12, minutes=33, seconds=1.234),
    "p_float": 1.345,
    "p_inet": ipaddress.ip_address("127.0.0.1"),
    "p_int": 1012,
    "p_smallint": 12,
    "p_text": "Àäxxyxy",
    "p_text_nulled": None,
    "p_text_omitted": None,
    "p_time": datetime.time(11, 22, 33, 456),
    # datetime in Python's stdlib is lossy after milliseconds:
    "p_timestamp": datetime.datetime(
        2021, 11, 21, 15, 34, 44, 123000, tzinfo=datetime.timezone.utc
    ),
    "p_tinyint": 7,
    "p_varint": 10293,
    "p_uuid": UUID("01932c57-b8b7-8310-a84e-6d3fada3c525"),
    "p_vector": [-0.1, 0.2, -0.3],
    "p_list_int": [10, 11, -12],
    "p_map_text_text": {"x": "YYY", "p": "QQQ"},
    "p_set_int": {-43, 111, 109, 0},
    "p_double_minf": float("-Infinity"),
    "p_double_pinf": float("Infinity"),
    "p_float_nan": float("NaN"),
}

COMPOSITE_VECTOR_ROWS_N = 3
COMPOSITE_VECTOR_ROWS = [
    {
        "p_text": p_t,
        "p_int": p_i,
        "p_boolean": p_i % 2 == 0,
        "p_vector": DataAPIVector([p_i + {"A": 0.0, "B": 0.1}[p_t], 1, 0]),
    }
    for p_t in {"A", "B"}
    for p_i in range(COMPOSITE_VECTOR_ROWS_N)
]

ALLMAPS_STDLIB_ROW = {
    "id": "sl_row",
    "text_map": {
        "text key": "text value",
    },
    "ascii_map": {
        "ascii key": "ascii value",
    },
    "inet_map": {
        ipaddress.IPv4Address("192.168.1.1"): ipaddress.IPv4Address("10.0.0.1"),
    },
    "date_map": {
        datetime.date(2025, 1, 1): datetime.date(2026, 1, 1),
    },
    "duration_map": {
        "duration text key": datetime.timedelta(seconds=4800),
    },
    "time_map": {
        datetime.time(12, 0): datetime.time(23, 59, 59, 999999),
    },
    "timestamp_map": {
        datetime.datetime(
            2025, 1, 1, 12, 0, tzinfo=datetime.timezone.utc
        ): datetime.datetime(2026, 1, 1, 12, 0, tzinfo=datetime.timezone.utc),
    },
    "uuid_map": {
        UUID("123e4567-e89b-12d3-a456-426614174000"): UUID(
            "550e8400-e29b-41d4-a716-446655440000"
        ),
    },
    "tinyint_map": {
        5: 10,
    },
    "varint_map": {
        100000000: 200000000000,
    },
    "decimal_map": {
        Decimal("1.23"): Decimal("4.56"),
    },
    "double_map": {
        1.1: 2.2,
    },
    "int_map": {
        42: 99,
    },
    "smallint_map": {
        12: 34,
    },
    "bigint_map": {
        1: 1,
    },
    "boolean_map": {
        True: False,
    },
    "float_map": {
        3.14: 6.28,
    },
    "blob_map": {
        b"Hello World": b"Hello World",
    },
}
ALLMAPS_CUSTOMTYPES_ROW = {
    "id": "ct_row",
    "text_map": DataAPIMap(
        [
            ("text key", "text value"),
        ]
    ),
    "ascii_map": DataAPIMap(
        [
            ("ascii key", "ascii value"),
        ]
    ),
    "inet_map": DataAPIMap(
        [
            (ipaddress.ip_address("192.168.1.1"), ipaddress.ip_address("10.0.0.1")),
        ]
    ),
    "date_map": DataAPIMap(
        [
            (
                DataAPIDate.from_string("2025-01-01"),
                DataAPIDate.from_string("2026-01-01"),
            ),
        ]
    ),
    "duration_map": DataAPIMap(
        [
            ("duration text key", DataAPIDuration.from_c_string("1h20m")),
        ]
    ),
    "time_map": DataAPIMap(
        [
            (
                DataAPITime.from_string("12:00:00.000000000"),
                DataAPITime.from_string("23:59:59.999999999"),
            ),
        ]
    ),
    "timestamp_map": DataAPIMap(
        [
            (
                DataAPITimestamp.from_string("2025-01-01T12:00:00.000Z"),
                DataAPITimestamp.from_string("2026-01-01T12:00:00.000Z"),
            ),
        ]
    ),
    "uuid_map": DataAPIMap(
        [
            (
                UUID("123e4567-e89b-12d3-a456-426614174000"),
                UUID("550e8400-e29b-41d4-a716-446655440000"),
            ),
        ]
    ),
    "tinyint_map": DataAPIMap(
        [
            (5, 10),
        ]
    ),
    "varint_map": DataAPIMap(
        [
            (100000000, 200000000000),
        ]
    ),
    "decimal_map": DataAPIMap(
        [
            (Decimal("1.23"), Decimal("4.56")),
        ]
    ),
    "double_map": DataAPIMap(
        [
            (1.1, 2.2),
        ]
    ),
    "int_map": DataAPIMap(
        [
            (42, 99),
        ]
    ),
    "smallint_map": DataAPIMap(
        [
            (12, 34),
        ]
    ),
    "bigint_map": DataAPIMap(
        [
            (1, 1),
        ]
    ),
    "boolean_map": DataAPIMap(
        [
            (True, False),
        ]
    ),
    "float_map": DataAPIMap(
        [
            (3.14, 6.28),
        ]
    ),
    "blob_map": DataAPIMap(
        [
            (b"Hello World", b"Hello World"),
        ]
    ),
}
ALLMAPS_CUSTOMTYPES_EMPTY_ROW = {
    "id": "empty_maps_customtypes",
    "ascii_map": DataAPIMap([]),
    "bigint_map": DataAPIMap([]),
    "blob_map": DataAPIMap([]),
    "boolean_map": DataAPIMap([]),
    "date_map": DataAPIMap([]),
    "decimal_map": DataAPIMap([]),
    "double_map": DataAPIMap([]),
    "duration_map": DataAPIMap([]),
    "float_map": DataAPIMap([]),
    "inet_map": DataAPIMap([]),
    "int_map": DataAPIMap([]),
    "smallint_map": DataAPIMap([]),
    "text_map": DataAPIMap([]),
    "time_map": DataAPIMap([]),
    "timestamp_map": DataAPIMap([]),
    "tinyint_map": DataAPIMap([]),
    "uuid_map": DataAPIMap([]),
    "varint_map": DataAPIMap([]),
}
ALLMAPS_STDLIB_EMPTY_ROW = {
    "id": "empty_maps_stdlib",
    "ascii_map": {},
    "bigint_map": {},
    "blob_map": {},
    "boolean_map": {},
    "date_map": {},
    "decimal_map": {},
    "double_map": {},
    "duration_map": {},
    "float_map": {},
    "inet_map": {},
    "int_map": {},
    "smallint_map": {},
    "text_map": {},
    "time_map": {},
    "timestamp_map": {},
    "tinyint_map": {},
    "uuid_map": {},
    "varint_map": {},
}

UDT_NAME = "test_udt"
UDT_DEF0 = CreateTypeDefinition(
    fields={
        "f_float0": "float",
        "f_float1": "float",
        "f_float2": "float",
    }
)
UDT_ALTER_OP_1 = AlterTypeAddFields(fields={"f_text0": "text"})
UDT_ALTER_OP_2 = AlterTypeRenameFields(fields={"f_float0": "z_float0"})
UDT_DEF1 = CreateTypeDefinition(
    fields={
        "z_float0": "float",
        "f_float1": "float",
        "f_float2": "float",
        "f_text0": "text",
    }
)

WEIRD_UDT_BASE_TABLE_NAME = "udt_weird_base"
WEIRD_UDT_BASE_TYPE_NAME = "wrd_type_base"
WEIRD_UDT_BASE_INITIALIZE_STATEMENTS = [
    f"create type IF NOT EXISTS {WEIRD_UDT_BASE_TYPE_NAME} (a text, b int)",
    (
        f"create table IF NOT EXISTS {WEIRD_UDT_BASE_TABLE_NAME} (id text primary key, f_udt "
        f"frozen<{WEIRD_UDT_BASE_TYPE_NAME}>, "
        f"f_l_udt frozen<list<{WEIRD_UDT_BASE_TYPE_NAME}>>);"
    ),
]
WEIRD_UDT_BASE_CLOSE_STATEMENTS = [
    f"drop table {WEIRD_UDT_BASE_TABLE_NAME};",
    f"drop type {WEIRD_UDT_BASE_TYPE_NAME};",
]
WEIRD_BASE_DOCUMENT_PK = {"id": "the_doc"}
WEIRD_BASE_DOCUMENT = {
    "f_udt": DataAPIDictUDT({"a": "A_of_f_udt", "b": 10}),
    "f_l_udt": [DataAPIDictUDT({"a": "A_of_f_l_udt", "b": 12})],
    **WEIRD_BASE_DOCUMENT_PK,
}

UNSUPPORTED_UDT_NAME = "test_unsupported_udt"
UNSUPPORTED_UDT_CREATE = (
    f"CREATE TYPE IF NOT EXISTS {UNSUPPORTED_UDT_NAME} (area FLOAT, trees SET<FLOAT>);"
)
UNSUPPORTED_UDT_DROP = f"DROP TYPE IF EXISTS {UNSUPPORTED_UDT_NAME};"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.event_observers import (
    ObservableError,
    ObservableEvent,
    ObservableEventType,
    ObservableRequest,
    ObservableResponse,
    ObservableWarning,
    Observer,
)
from astrapy.exceptions import (
    DataAPIErrorDescriptor,
    DataAPIWarningDescriptor,
)

OBS_ERR = ObservableError(
    error=DataAPIErrorDescriptor(
        {
            "title": "error_title",
            "errorCode": "error_errorCode",
            "message": "error_message",
        }
    ),
)
OBS_WRN = ObservableWarning(
    warning=DataAPIWarningDescriptor(
        {
            "title": "warning_title",
            "errorCode": "warning_errorCode",
            "message": "warning_message",
        }
    ),
)
OBS_RSP_1 = ObservableResponse(body='{"k_resp":"v1"}', status_code=123)
OBS_RSP_2 = ObservableResponse(body='{"k_resp":"v2"}', status_code=123)
OBS_REQ = ObservableRequest(
    payload='{"k_req":"v"}',
    http_method="PUT",
    url="http://bla.bla",
    query_parameters={"k": 10},
    redacted_headers={"x-aaa": "xyz"},
    dev_ops_api=False,
)


class TestEventObservers:
    @pytest.mark.describe("test of custom observer receiving events")
    def test_custom_observer(self) -> None:
        class MyObserver(Observer):
            def __init__(
                self, evt_map: dict[ObservableEventType, list[ObservableEvent]]
            ) -> None:
                self.evt_map = evt_map

            def receive(
                self,
                event: ObservableEvent,
                sender: Any = None,
                function_name: str | None = None,
                request_id: str | None = None,
            ) -> None:
                self.evt_map[event.event_type] = self.evt_map.get(
                    event.event_type, []
                ) + [event]

        received_events: dict[ObservableEventType, list[ObservableEvent]] = {}
        my_obs = MyObserver(received_events)

        my_obs.receive(OBS_RSP_1)
        my_obs.receive(OBS_ERR)
        my_obs.receive(OBS_WRN)
        my_obs.receive(OBS_RSP_2)
        my_obs.receive(OBS_REQ)

        assert received_events[ObservableEventType.ERROR] == [OBS_ERR]
        assert received_events[ObservableEventType.WARNING] == [OBS_WRN]
        assert received_events[ObservableEventType.RESPONSE] == [OBS_RSP_1, OBS_RSP_2]
        assert received_events[ObservableEventType.REQUEST] == [OBS_REQ]
        assert len(received_events) == 4

    @pytest.mark.describe("test of observer from event dict")
    def test_observer_from_evdict(self) -> None:
        received_events: dict[ObservableEventType, list[ObservableEvent]] = {}
        my_obs = Observer.from_event_dict(received_events)

        my_obs.receive(OBS_RSP_1)
        my_obs.receive(OBS_ERR)
        my_obs.receive(OBS_WRN)
        my_obs.receive(OBS_RSP_2)
        my_obs.receive(OBS_REQ)

        assert received_events[ObservableEventType.ERROR] == [OBS_ERR]
        assert received_events[ObservableEventType.WARNING] == [OBS_WRN]
        assert received_events[ObservableEventType.RESPONSE] == [OBS_RSP_1, OBS_RSP_2]
        assert received_events[ObservableEventType.REQUEST] == [OBS_REQ]
        assert len(received_events) == 4

        # now with filtering by event type
        received_events_f: dict[ObservableEventType, list[ObservableEvent]] = {}
        my_obs_f = Observer.from_event_dict(
            received_events_f,
            event_types=(ObservableEventType.RESPONSE, ObservableEventType.WARNING),
        )

        my_obs_f.receive(OBS_RSP_1)
        my_obs_f.receive(OBS_ERR)
        my_obs_f.receive(OBS_WRN)
        my_obs_f.receive(OBS_RSP_2)
        my_obs_f.receive(OBS_REQ)

        assert received_events_f[ObservableEventType.RESPONSE] == [
            OBS_RSP_1,
            OBS_RSP_2,
        ]
        assert received_events_f[ObservableEventType.WARNING] == [OBS_WRN]
        assert len(received_events_f) == 2

    @pytest.mark.describe("test of observer from event list")
    def test_observer_from_evlist(self) -> None:
        ev_list: list[ObservableEvent] = []
        my_obs = Observer.from_event_list(ev_list)

        my_obs.receive(OBS_RSP_1)
        my_obs.receive(OBS_ERR)
        my_obs.receive(OBS_WRN)
        my_obs.receive(OBS_RSP_2)
        my_obs.receive(OBS_REQ)

        assert ev_list == [
            OBS_RSP_1,
            OBS_ERR,
            OBS_WRN,
            OBS_RSP_2,
            OBS_REQ,
        ]

        # now with filtering by event type
        ev_list_f: list[ObservableEvent] = []
        my_obs_f = Observer.from_event_list(
            ev_list_f,
            event_types=(ObservableEventType.RESPONSE, ObservableEventType.WARNING),
        )

        my_obs_f.receive(OBS_RSP_1)
        my_obs_f.receive(OBS_ERR)
        my_obs_f.receive(OBS_WRN)
        my_obs_f.receive(OBS_RSP_2)
        my_obs_f.receive(OBS_REQ)

        assert ev_list_f == [
            OBS_RSP_1,
            OBS_WRN,
            OBS_RSP_2,
        ]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest
from deprecation import DeprecatedWarning

from astrapy.data.info.database_info import (
    AstraDBAdminDatabaseRegionInfo,
    AstraDBAvailableRegionInfo,
)

from ..conftest import is_future_version


class TestRegionNameDeprecation:
    @pytest.mark.describe("test of region_name in AstraDBAdminDatabaseRegionInfo")
    def test_regionname_dbregioninfo(self) -> None:
        dc0 = {
            "capacityUnits": 1,
            "cloudProvider": "AWS",
            "dateCreated": "2025-04-10T19:14:44Z",
            "id": "...",
            "isPrimary": True,
            "name": "dc-1",
            "region": "us-east-2",
            "regionClassification": "standard",
            "regionZone": "na",
            "requestedNodeCount": 3,
            "secureBundleInternalUrl": "...",
            "secureBundleMigrationProxyInternalUrl": "...",
            "secureBundleMigrationProxyUrl": "...",
            "secureBundleUrl": "...",
            "status": "",
            "streamingTenant": {},
            "targetAccount": "abcd0123",
            "tier": "serverless",
        }
        region_info = AstraDBAdminDatabaseRegionInfo(
            raw_datacenter_dict=dc0,
            environment="dev",
            database_id="D",
        )

        r0 = region_info.name
        with pytest.warns(DeprecationWarning) as w_checker:
            r1 = region_info.region_name
        assert len(w_checker.list) == 1
        warning0 = w_checker.list[0].message
        assert isinstance(warning0, DeprecatedWarning)
        assert is_future_version(warning0.removed_in)

        assert r0 == r1

    @pytest.mark.describe("test of region_name in AstraDBAvailableRegionInfo")
    def test_regionname_availableregion(self) -> None:
        ar0 = {
            "classification": "premium",
            "cloudProvider": "GCP",
            "displayName": "West Europe3 (Frankfurt, Germany)",
            "enabled": True,
            "name": "europe-west3",
            "region_type": "serverless",
            "reservedForQualifiedUsers": True,
            "zone": "emea",
        }
        available_region_info = AstraDBAvailableRegionInfo._from_dict(ar0)

        r0 = available_region_info.name
        with pytest.warns(DeprecationWarning) as w_checker:
            r1 = available_region_info.region_name

        assert len(w_checker.list) == 1
        warning0 = w_checker.list[0].message
        assert isinstance(warning0, DeprecatedWarning)
        assert is_future_version(warning0.removed_in)

        assert r0 == r1



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.info import ListTypeDescriptor


class TestListTypeDescriptor:
    @pytest.mark.describe("test of ListTypeDescriptor parsing and back")
    def test_listtypedescriptor_parsing(self) -> None:
        types_json = [
            {
                "type": "userDefined",
                "udtName": "tame_type",
                "definition": {
                    "fields": {
                        "t": {
                            "type": "text",
                        },
                        "i": {
                            "type": "int",
                        },
                    },
                },
                "apiSupport": {
                    "createTable": True,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "default_keyspace.tame_type",
                },
            },
            {
                "type": "userDefined",
                "udtName": "lazy_type",
                "definition": {
                    "fields": {
                        "b": {
                            "type": "blob",
                        },
                        "a": {
                            "type": "ascii",
                        },
                    },
                },
            },
        ]

        for type_json in types_json:
            t_fd = ListTypeDescriptor._from_dict(type_json)
            t_co = ListTypeDescriptor.coerce(type_json)

            assert t_fd == t_co

            assert t_fd.as_dict() == type_json

    @pytest.mark.describe("test of ListTypeDescriptor, faulty and unsupported UDTs")
    def test_listtypedescriptor_weird_udts(self) -> None:
        # good; empty; unsupported
        good_dict = {
            "type": "userDefined",
            "udtName": "bla",
            "definition": {"fields": {"a": {"type": "int"}}},
            "apiSupport": {
                "createTable": True,
                "insert": True,
                "read": True,
                "filter": False,
                "cqlDefinition": "default_keyspace.bla",
            },
        }
        empty_dict: dict[str, Any] = {}
        unsupported_dict = {
            "type": "UNSUPPORTED",
            "apiSupport": {
                "createTable": False,
                "insert": False,
                "read": False,
                "filter": False,
                "cqlDefinition": "default_keyspace.blu",
            },
        }

        assert ListTypeDescriptor._is_valid_dict(good_dict)
        assert not ListTypeDescriptor._is_valid_dict(empty_dict)
        assert ListTypeDescriptor._is_valid_dict(unsupported_dict)

        good_parsed = ListTypeDescriptor._from_dict(good_dict)
        assert good_parsed.as_dict() == good_dict

        with pytest.raises(KeyError):
            ListTypeDescriptor._from_dict(empty_dict)

        unsupported_parsed = ListTypeDescriptor._from_dict(unsupported_dict)
        assert unsupported_parsed.as_dict() == unsupported_dict



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    TypeVar,
    Union,
)

from astrapy.data_types import DataAPIVector

if TYPE_CHECKING:
    from astrapy.info import CreateTypeDefinition


DefaultDocumentType = dict[str, Any]
DefaultRowType = dict[str, Any]
ProjectionType = Union[
    Iterable[str], dict[str, Union[bool, dict[str, Union[int, Iterable[int]]]]]
]
SortType = dict[str, Any]
HybridSortType = dict[
    str, Union[str, dict[str, Union[str, list[float], DataAPIVector]]]
]
FilterType = dict[str, Any]
CallerType = tuple[Optional[str], Optional[str]]
SerializerFunctionType = Callable[[Any], dict[str, Any]]
UDTDeserializerFunctionType = Callable[
    [dict[str, Any], Optional["CreateTypeDefinition"]], Any
]

ROW = TypeVar("ROW")
ROW2 = TypeVar("ROW2")
DOC = TypeVar("DOC")
DOC2 = TypeVar("DOC2")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.info import (
    CollectionDefinition,
    CollectionDescriptor,
    CollectionLexicalOptions,
    CollectionRerankOptions,
    RerankServiceOptions,
)


class TestFindAndRerankCollectionDefinition:
    @pytest.mark.describe("test of FARR collection definitions conversions")
    def test_farr_collectiondefinition_conversions(self) -> None:
        api_coll_descs: list[dict[str, Any]] = [
            {
                "name": "col_name",
                "options": {
                    "lexical": {
                        "enabled": False,
                        "analyzer": "the_analyzer",
                    },
                },
            },
            {
                "name": "col_name",
                "options": {
                    "rerank": {
                        "enabled": True,
                        "service": {
                            "provider": "the_pro",
                            "modelName": "the_mo",
                        },
                    },
                },
            },
        ]

        for cd_dict in api_coll_descs:
            descriptor = CollectionDescriptor._from_dict(cd_dict)
            # dict->obj->dict test
            assert descriptor.as_dict() == cd_dict
            if "options" in cd_dict:
                cdef_dict = cd_dict["options"]
                assert CollectionDefinition._from_dict(cdef_dict).as_dict() == cdef_dict
            # obj->dict->obj test
            assert CollectionDescriptor._from_dict(descriptor.as_dict()) == descriptor
            definition = descriptor.definition
            assert CollectionDefinition._from_dict(definition.as_dict()) == definition
            # coerce calls
            assert CollectionDescriptor.coerce(cd_dict) == descriptor
            assert CollectionDescriptor.coerce(descriptor) == descriptor
            if "options" in cd_dict:
                assert CollectionDefinition.coerce(cd_dict["options"]) == definition
            assert CollectionDefinition.coerce(definition) == definition

    @pytest.mark.describe("test of FARR collection lexical options conversions")
    def test_farr_collectionlexicaloptions_conversions(self) -> None:
        lx1 = CollectionLexicalOptions(analyzer="a")
        assert lx1.as_dict() == {"analyzer": "a", "enabled": True}
        assert CollectionLexicalOptions._from_dict(lx1.as_dict()) == lx1

        lx2 = CollectionLexicalOptions(analyzer="b", enabled=True)
        assert lx2.as_dict() == {"analyzer": "b", "enabled": True}
        assert CollectionLexicalOptions._from_dict(lx2.as_dict()) == lx2

        lx3 = CollectionLexicalOptions(analyzer={"x": 1}, enabled=False)
        assert lx3.as_dict() == {"analyzer": {"x": 1}, "enabled": False}
        assert CollectionLexicalOptions._from_dict(lx3.as_dict()) == lx3

    @pytest.mark.describe("test of FARR collection rerank options conversions")
    def test_farr_collectionrerankoptions_conversions(self) -> None:
        sv1 = RerankServiceOptions(
            provider="provider",
            model_name="model_name",
            authentication={"a": "z"},
            parameters={"p": "q"},
        )
        assert sv1.as_dict() == {
            "provider": "provider",
            "modelName": "model_name",
            "authentication": {"a": "z"},
            "parameters": {"p": "q"},
        }
        assert RerankServiceOptions._from_dict(sv1.as_dict()) == sv1

        sv2 = RerankServiceOptions(
            provider="pr",
            model_name="mn",
        )
        assert sv2.as_dict() == {"provider": "pr", "modelName": "mn"}
        assert RerankServiceOptions._from_dict(sv2.as_dict()) == sv2

        rr1 = CollectionRerankOptions(service=sv1)
        assert rr1.as_dict() == {"service": sv1.as_dict(), "enabled": True}
        assert CollectionRerankOptions._from_dict(rr1.as_dict()) == rr1

        rr2 = CollectionRerankOptions(service=None, enabled=True)
        assert rr2.as_dict() == {"enabled": True}
        assert CollectionRerankOptions._from_dict(rr2.as_dict()) == rr2

        rr3 = CollectionRerankOptions(service=sv2, enabled=False)
        assert rr3.as_dict() == {"service": sv2.as_dict(), "enabled": False}
        assert CollectionRerankOptions._from_dict(rr3.as_dict()) == rr3

    @pytest.mark.describe(
        "test of FARR collection definition, rerank builder interface"
    )
    def test_farr_collectiondefinition_rerank_builders(self) -> None:
        zero = CollectionDefinition.builder()

        rer1 = zero.set_rerank("p", "m", parameters={"p": "q"}).build()
        assert rer1.as_dict() == {
            "rerank": {
                "enabled": True,
                "service": {
                    "provider": "p",
                    "modelName": "m",
                    "parameters": {"p": "q"},
                },
            }
        }

        r_svc = RerankServiceOptions("P", "M")
        rer2 = zero.set_rerank(r_svc).build()
        assert rer2.as_dict() == {
            "rerank": {
                "enabled": True,
                "service": {"provider": "P", "modelName": "M"},
            }
        }

        cr_opt = CollectionRerankOptions(
            service=RerankServiceOptions("pp", "mm"),
        )
        rer3 = zero.set_rerank(cr_opt).build()
        assert rer3.as_dict() == {
            "rerank": {
                "enabled": True,
                "service": {"provider": "pp", "modelName": "mm"},
            }
        }

        rer_z = zero.set_rerank("x", "y")
        reset = rer_z.set_rerank(None).build()
        assert reset.as_dict() == {}

    @pytest.mark.describe(
        "test of FARR collection definition, lexical builder interface"
    )
    def test_farr_collectiondefinition_lexical_builders(self) -> None:
        zero = CollectionDefinition.builder()

        lex1 = zero.set_lexical("an").build()
        assert lex1.as_dict() == {
            "lexical": {
                "enabled": True,
                "analyzer": "an",
            }
        }

        lx_opt = CollectionLexicalOptions(enabled=False, analyzer="AN")
        lex2 = zero.set_lexical(lx_opt).build()
        assert lex2.as_dict() == {
            "lexical": {
                "enabled": False,
                "analyzer": "AN",
            }
        }

        lex3 = zero.set_lexical(enabled=False)
        assert lex3.as_dict() == {"lexical": {"enabled": False}}

        lex4 = zero.set_lexical(enabled=True)
        assert lex4.as_dict() == {"lexical": {"enabled": True}}

        lex_z = zero.set_lexical("aaa")
        reset = lex_z.set_lexical(None)
        assert reset.as_dict() == {}



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Any

from astrapy.exceptions.error_descriptors import (
    DataAPIErrorDescriptor,
    DataAPIWarningDescriptor,
)
from astrapy.utils.str_enum import StrEnum


class ObservableEventType(StrEnum):
    """
    Enum for the possible values of the event type for observable events
    """

    WARNING = "warning"
    ERROR = "error"
    REQUEST = "request"
    RESPONSE = "response"


@dataclass
class ObservableEvent(ABC):
    """
    Class that represents the most general 'event' that is sent to observers.

    Attributes:
        event_type: the type of the event, such as "log", "error", or "warning".
    """

    event_type: ObservableEventType


@dataclass
class ObservableError(ObservableEvent):
    """
    An event representing an error returned from the Data API in a response.

    These are dispatched unconditionally to the attached observers as the
    response is parsed. The actual raising of an exception does not always
    follow; moreover, further operations may take place before that occurs.

    Note:
        Only errors returned within the Data API response in the
        "errors" field are dispatched this way. The most general exception that
        can occur during a method call are not necessarily of this form.

    Attributes:
        event_type: it has value ObservableEventType.ERROR in this case.
        error: a descriptor of the error, as found in the Data API response.
    """

    error: DataAPIErrorDescriptor

    def __init__(self, error: DataAPIErrorDescriptor) -> None:
        self.event_type = ObservableEventType.ERROR
        self.error = error


@dataclass
class ObservableWarning(ObservableEvent):
    """
    An event representing a warning returned by a Data API command.

    These are dispatched to the attached observers as the response is parsed.

    Attributes:
        event_type: it has value ObservableEventType.WARNING in this case.
        warning: a descriptor of the warning, as found in the Data API response.
    """

    warning: DataAPIWarningDescriptor

    def __init__(self, warning: DataAPIWarningDescriptor) -> None:
        self.event_type = ObservableEventType.WARNING
        self.warning = warning


@dataclass
class ObservableRequest(ObservableEvent):
    """
    An event representing a request being sent, captured with its
    payload exactly as will be sent to the API.

    Attributes:
        event_type: it has value ObservableEventType.REQUEST in this case.
        payload: the payload as a string.
        http_method: one of `astrapy.utils.request_tools.HttpMethod`, e.g. "POST".
        url: the complete URL the request is targeted at.
        query_parameters: if present, all query parameters in dict form.
        redacted_headers: a dictionary of the non-sensitive headers being used
            for the request. Authentication credentials and API Keys are removed.
        dev_ops_api: true if and only if the request is aimed at the DevOps API.
    """

    payload: str | None
    http_method: str
    url: str
    query_parameters: dict[str, Any] | None
    redacted_headers: dict[str, Any] | None
    dev_ops_api: bool

    def __init__(
        self,
        payload: str | None,
        http_method: str,
        url: str,
        query_parameters: dict[str, Any] | None,
        redacted_headers: dict[str, Any] | None,
        dev_ops_api: bool,
    ) -> None:
        self.event_type = ObservableEventType.REQUEST
        self.payload = payload
        self.http_method = http_method
        self.url = url
        self.query_parameters = query_parameters
        self.redacted_headers = redacted_headers
        self.dev_ops_api = dev_ops_api


@dataclass
class ObservableResponse(ObservableEvent):
    """
    An event representing a response received by the Data API, whose body
    is captured exactly as is sent by the Data API.

    Attributes:
        event_type: it has value ObservableEventType.RESPONSE in this case.
        body: a string expressing the response body.
        status_code: the response HTTP status code.
    """

    body: str | None
    status_code: int

    def __init__(self, body: str | None, *, status_code: int) -> None:
        self.event_type = ObservableEventType.RESPONSE
        self.body = body
        self.status_code = status_code



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime

import pytest

from astrapy.data.info.table_descriptor.table_columns import TableColumnTypeDescriptor
from astrapy.data.utils.table_converters import (
    create_row_tpostprocessor,
    preprocess_table_payload,
)
from astrapy.data_types import DataAPITimestamp
from astrapy.info import (
    ColumnType,
    TableScalarColumnTypeDescriptor,
)
from astrapy.utils.api_options import SerdesOptions, defaultSerdesOptions


class TestDatetimeSerdesOptions:
    @pytest.mark.describe("test reading timestamp from table, custom dt")
    def test_ddate_table_reading_customdt(self) -> None:
        tz = datetime.timezone(datetime.timedelta(hours=2, minutes=45))
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(
                custom_datatypes_in_reading=True,
                datetime_tzinfo=tz,
            ),
        )
        columns: dict[str, TableColumnTypeDescriptor] = {
            "mu": TableScalarColumnTypeDescriptor(ColumnType.TIMESTAMP),
        }
        postprocessor = create_row_tpostprocessor(
            columns=columns,
            options=sd_options,
            similarity_pseudocolumn=None,
        )

        ts = DataAPITimestamp.from_string("1991-07-23T12:34:56+01:30")
        raw_response = {"mu": ts.to_string()}
        postprocessed = postprocessor(raw_response)
        assert postprocessed["mu"] == ts

    @pytest.mark.describe("test reading timestamp from table, stdlib and tzaware")
    def test_ddate_table_reading_stdlibtz(self) -> None:
        tz = datetime.timezone(datetime.timedelta(hours=2, minutes=45))
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(
                custom_datatypes_in_reading=False,
                datetime_tzinfo=tz,
            ),
        )
        columns: dict[str, TableColumnTypeDescriptor] = {
            "mu": TableScalarColumnTypeDescriptor(ColumnType.TIMESTAMP),
        }
        postprocessor = create_row_tpostprocessor(
            columns=columns,
            options=sd_options,
            similarity_pseudocolumn=None,
        )

        dt = datetime.datetime(1991, 7, 23, 12, 34, 56, tzinfo=tz)
        raw_response = {"mu": DataAPITimestamp.from_datetime(dt).to_string()}
        postprocessed = postprocessor(raw_response)
        assert postprocessed["mu"] == dt

    @pytest.mark.describe("test reading timestamp from table, stdlib and naive")
    def test_ddate_table_reading_stdlibnaive(self) -> None:
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(
                custom_datatypes_in_reading=False,
                datetime_tzinfo=None,
            ),
        )
        columns: dict[str, TableColumnTypeDescriptor] = {
            "mu": TableScalarColumnTypeDescriptor(ColumnType.TIMESTAMP),
        }
        postprocessor = create_row_tpostprocessor(
            columns=columns,
            options=sd_options,
            similarity_pseudocolumn=None,
        )

        dt = datetime.datetime(1991, 7, 23, 12, 34, 56)
        raw_response = {"mu": DataAPITimestamp.from_datetime(dt).to_string()}
        postprocessed = postprocessor(raw_response)
        assert postprocessed["mu"] == dt

    @pytest.mark.describe("test writing tzaware datetime to table, no naive allowed")
    def test_ddate_table_writing_tzaware_strict(self) -> None:
        tz = datetime.timezone(datetime.timedelta(hours=2, minutes=45))
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(accept_naive_datetimes=False),
        )
        dt = datetime.datetime(1991, 7, 23, 12, 34, 56, tzinfo=tz)
        preprocessed = preprocess_table_payload(
            {"mu": dt},
            options=sd_options,
            map2tuple_checker=None,
        )
        assert preprocessed is not None
        assert DataAPITimestamp.from_string(
            preprocessed["mu"]
        ) == DataAPITimestamp.from_datetime(dt)

    @pytest.mark.describe("test writing naive datetime to table, no naive allowed")
    def test_ddate_table_writing_naive_strict(self) -> None:
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(accept_naive_datetimes=False),
        )
        dt = datetime.datetime(1991, 7, 23, 12, 34, 56)
        with pytest.raises(ValueError, match="tz"):
            preprocess_table_payload(
                {"mu": dt},
                options=sd_options,
                map2tuple_checker=None,
            )

    @pytest.mark.describe("test writing tzaware datetime to table, naive permitted")
    def test_ddate_table_writing_tzaware_relaxed(self) -> None:
        tz = datetime.timezone(datetime.timedelta(hours=2, minutes=45))
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(accept_naive_datetimes=True),
        )
        dt = datetime.datetime(1991, 7, 23, 12, 34, 56, tzinfo=tz)
        preprocessed = preprocess_table_payload(
            {"mu": dt},
            options=sd_options,
            map2tuple_checker=None,
        )
        assert preprocessed is not None
        assert DataAPITimestamp.from_string(
            preprocessed["mu"]
        ) == DataAPITimestamp.from_datetime(dt)

    @pytest.mark.describe("test writing naive datetime to table, naive permitted")
    def test_ddate_table_writing_naive_relaxed(self) -> None:
        sd_options = defaultSerdesOptions.with_override(
            SerdesOptions(accept_naive_datetimes=True),
        )
        dt = datetime.datetime(1991, 7, 23, 12, 34, 56)
        preprocessed = preprocess_table_payload(
            {"mu": dt},
            options=sd_options,
            map2tuple_checker=None,
        )
        assert preprocessed is not None
        assert DataAPITimestamp.from_string(
            preprocessed["mu"]
        ) == DataAPITimestamp.from_datetime(dt)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
from typing import cast


def get_python_version() -> tuple[int, int, int]:
    """
    Get the version of the currently-running Python interpreter.

    Returns:
        version: (major, minor, revision) as a tuple[int, int, int]
    """

    return cast(tuple[int, int, int], tuple(sys.version_info[:3]))



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the parsing of API endpoints and related
"""

from __future__ import annotations

import pytest

from astrapy.admin import ParsedAPIEndpoint, parse_api_endpoint
from astrapy.info import AstraDBAvailableRegionInfo


@pytest.mark.describe("test of parsing API endpoints")
def test_parse_api_endpoint() -> None:
    parsed_prod = parse_api_endpoint(
        "https://01234567-89ab-cdef-0123-456789abcdef-eu-west-1.apps.astra.datastax.com"
    )
    assert isinstance(parsed_prod, ParsedAPIEndpoint)
    assert parsed_prod == ParsedAPIEndpoint(
        database_id="01234567-89ab-cdef-0123-456789abcdef",
        region="eu-west-1",
        environment="prod",
    )

    parsed_prod = parse_api_endpoint(
        "https://a1234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-dev.datastax.com"
    )
    assert isinstance(parsed_prod, ParsedAPIEndpoint)
    assert parsed_prod == ParsedAPIEndpoint(
        database_id="a1234567-89ab-cdef-0123-456789abcdef",
        region="us-central1",
        environment="dev",
    )

    parsed_prod = parse_api_endpoint(
        "https://b1234567-89ab-cdef-0123-456789abcdef-eu-southwest-4.apps.astra-test.datastax.com/subpath?a=1"
    )
    assert isinstance(parsed_prod, ParsedAPIEndpoint)
    assert parsed_prod == ParsedAPIEndpoint(
        database_id="b1234567-89ab-cdef-0123-456789abcdef",
        region="eu-southwest-4",
        environment="test",
    )

    malformed_endpoints = [
        "http://01234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-dev.datastax.com",
        "https://a909bdbf-q9ba-4e5e-893c-a859ed701407-us-central1.apps.astra-dev.datastax.com",
        "https://01234567-89ab-cdef-0123-456789abcdef-us-c_entral1.apps.astra-dev.datastax.com",
        "https://01234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-fake.datastax.com",
        "https://01234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-dev.datastax-staging.com",
        "ahttps://01234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-dev.datastax.com",
    ]

    for m_ep in malformed_endpoints:
        assert parse_api_endpoint(m_ep) is None


@pytest.mark.describe("test of marshaling of available region info")
def test_parse_availableregioninfo() -> None:
    region_dict = {
        "classification": "standard",
        "cloudProvider": "AWS",
        "displayName": "US East (Ohio)",
        "enabled": True,
        "name": "us-east-2",
        "region_type": "vector",
        "reservedForQualifiedUsers": False,
        "zone": "na",
    }
    assert AstraDBAvailableRegionInfo._from_dict(region_dict).as_dict() == region_dict



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio

# TODO NOINETINUDT
# import ipaddress
from typing import TYPE_CHECKING, Any

import pytest

from astrapy import AsyncDatabase
from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import SerializerFunctionType, UDTDeserializerFunctionType
from astrapy.data.utils.table_types import TableUnsupportedColumnType
from astrapy.data_types import (
    DataAPIDictUDT,
    DataAPIMap,
    DataAPISet,
)
from astrapy.exceptions import DataAPIResponseException

from ..conftest import (
    CQL_AVAILABLE,
    EXTENDED_PLAYER_TYPE_NAME,
    PLAYER_TYPE_NAME,
    # THE_BYTES,
    THE_DATETIME,
    THE_TIMESTAMP,
    THE_TIMEZONE,
    DataAPICredentials,
    DefaultAsyncTable,
    ExtendedPlayer,
    NullablePlayer,
    Player,
    _extended_player_from_dict,
    _extended_player_serializer,
    _nullable_player_from_dict,
    _nullable_player_serializer,
    _player_from_dict,
    _player_serializer,
    dict_equal_same_class,
)
from .table_cql_assets import _extract_udt_definition
from .table_row_assets import (
    UDT_ALTER_OP_1,
    UDT_ALTER_OP_2,
    UDT_DEF0,
    UDT_DEF1,
    UDT_NAME,
    UNSUPPORTED_UDT_CREATE,
    UNSUPPORTED_UDT_DROP,
    UNSUPPORTED_UDT_NAME,
    WEIRD_BASE_DOCUMENT,
    WEIRD_BASE_DOCUMENT_PK,
    WEIRD_UDT_BASE_CLOSE_STATEMENTS,
    WEIRD_UDT_BASE_INITIALIZE_STATEMENTS,
    WEIRD_UDT_BASE_TABLE_NAME,
)

if TYPE_CHECKING:
    from cassandra.cluster import Session

SERIALIZER_BY_CLASS: dict[type, SerializerFunctionType] = {
    ExtendedPlayer: _extended_player_serializer,
    NullablePlayer: _nullable_player_serializer,
    Player: _player_serializer,
}


@pytest.mark.skipif(not CQL_AVAILABLE, reason="No CQL session available")
class TestTableUserDefinedTypes:
    @pytest.mark.describe("Test of UDT lifecycle, async")
    async def test_table_udt_lifecycle_async(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        cql_session: Session,
        async_database: AsyncDatabase,
    ) -> None:
        try:
            await async_database.create_type(UDT_NAME, definition=UDT_DEF0)
            with pytest.raises(DataAPIResponseException):
                await async_database.create_type(UDT_NAME, definition=UDT_DEF0)
            await async_database.create_type(
                UDT_NAME,
                definition=UDT_DEF0,
                if_not_exists=True,
            )

            assert UDT_DEF0 == _extract_udt_definition(
                cql_session,
                data_api_credentials_kwargs["keyspace"],
                UDT_NAME,
            )

            assert UDT_NAME in (await async_database.list_type_names())
            db_types_match = [
                t_def
                for t_def in (await async_database.list_types())
                if t_def.udt_name == UDT_NAME
            ]
            assert db_types_match != []
            assert db_types_match[0].definition == UDT_DEF0

            await async_database.alter_type(UDT_NAME, operation=UDT_ALTER_OP_1)
            await async_database.alter_type(UDT_NAME, operation=UDT_ALTER_OP_2)
            with pytest.raises(DataAPIResponseException):
                await async_database.alter_type(UDT_NAME, operation=UDT_ALTER_OP_1)

            assert UDT_DEF1 == _extract_udt_definition(
                cql_session,
                data_api_credentials_kwargs["keyspace"],
                UDT_NAME,
            )

            await async_database.drop_type(UDT_NAME)
            assert (
                _extract_udt_definition(
                    cql_session,
                    data_api_credentials_kwargs["keyspace"],
                    UDT_NAME,
                )
                is None
            )
            with pytest.raises(DataAPIResponseException):
                await async_database.drop_type(UDT_NAME)
            await async_database.drop_type(UDT_NAME, if_exists=True)

            await async_database.create_type(UDT_NAME, definition=UDT_DEF0)
            await async_database.drop_type(UDT_NAME)
        finally:
            await async_database.drop_type(UDT_NAME, if_exists=True)

    @pytest.mark.parametrize(
        ("encode_maps_as_lists_in_tables",),
        [
            ("NEVER",),
            ("DATAAPIMAPS",),
            ("ALWAYS",),
        ],
        ids=[
            "tuplesNEVER",
            "tuplesDATAAPIMAPS",
            "tuplesALWAYS",
        ],
    )
    @pytest.mark.parametrize(
        ("udt_format",),
        [
            ("dict",),
            ("dictwrapper",),
            ("dataclass",),
        ],
        ids=[
            "dict",
            "dictWrapper",
            "dataclass",
        ],
    )
    @pytest.mark.parametrize(
        ("udt_mode",),
        [
            ("simple",),
            ("partial",),
            ("empty",),
            ("extended",),
        ],
        ids=[
            "nonNullable",
            "partialNullable",
            "emptyNullable",
            "extendedClass",
        ],
    )
    @pytest.mark.describe("Test of UDT DML, async")
    async def test_table_udt_dml_async(
        self,
        player_udt: str,
        extended_player_udt: str,
        async_empty_table_udt_player: DefaultAsyncTable,
        async_empty_table_udt_extended_player: DefaultAsyncTable,
        encode_maps_as_lists_in_tables: str,
        udt_format: str,
        udt_mode: str,
    ) -> None:
        """
        This is a test of reading/writing UDTs on tables, within various column types.
        The test is heavily parametrized. Each parameter is a different 'write mode',
        but each individual test runs the read with all relevant read modes.
        'Write modes' = serdes options, the type used to express the UDT, the content.

        Note all "udt_mode" use one table except the "extended" (different UDT schema).
        The scenario of UDT as plain dicts + always maps as lists of pairs is forbidden,
        hence skipped.

        So the count of tested param combinations is as follows:
            simple UDT (2 fields)               extended UDT (4 fields) | serdesMaps:
            ------------------------------------------------------------+------------
             dict: 6  | nondict: 12              dict: 2  | nondict:  4 | nonALWAYS
            [dict: 3] | nondict: 6              [dict: 1] | nondict:  2 |    ALWAYS
            ------------------------------------------------------------+------------
        """

        # combinations of write settings that are not supposed to work are skipped:
        if udt_format == "dict" and encode_maps_as_lists_in_tables == "ALWAYS":
            pytest.skip("The Data API is not supposed to accept such a write format.")

        # choice of serdes options for writes
        src_atable: DefaultAsyncTable
        if udt_mode == "extended":
            src_atable = async_empty_table_udt_extended_player
        else:
            src_atable = async_empty_table_udt_player
        atable = src_atable.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(
                    serializer_by_class=dict(SERIALIZER_BY_CLASS),
                    encode_maps_as_lists_in_tables=encode_maps_as_lists_in_tables,
                ),
            ),
        )

        # asset preparation
        udt_unit: Any
        if udt_mode == "simple":
            if udt_format == "dict":
                udt_unit = {
                    "name": "John",
                    "age": 90,
                }
            elif udt_format == "dictwrapper":
                udt_unit = DataAPIDictUDT(
                    {
                        "name": "John",
                        "age": 90,
                    }
                )
            elif udt_format == "dataclass":
                udt_unit = Player(
                    name="John",
                    age=90,
                )
            else:
                raise ValueError(f"Unknown udt format '{udt_format}'.")
        elif udt_mode == "partial":
            if udt_format == "dict":
                udt_unit = {
                    "age": 90,
                }
            elif udt_format == "dictwrapper":
                udt_unit = DataAPIDictUDT(
                    {
                        "age": 90,
                    }
                )
            elif udt_format == "dataclass":
                udt_unit = NullablePlayer(
                    age=90,
                )
            else:
                raise ValueError(f"Unknown udt format '{udt_format}'.")
        elif udt_mode == "empty":
            if udt_format == "dict":
                udt_unit = {}
            elif udt_format == "dictwrapper":
                udt_unit = DataAPIDictUDT({})
            elif udt_format == "dataclass":
                udt_unit = NullablePlayer()
            else:
                raise ValueError(f"Unknown udt format '{udt_format}'.")
        elif udt_mode == "extended":
            if udt_format == "dict":
                udt_unit = {
                    "name": "John",
                    "age": 90,
                    # TODO NOINETINUDT
                    # "inet": ipaddress.ip_address("127.0.0.1"),
                    # TODO NOBLOBINUDT
                    # "blb": THE_BYTES,
                    "ts": THE_TIMESTAMP,
                }
            elif udt_format == "dictwrapper":
                udt_unit = DataAPIDictUDT(
                    {
                        "name": "John",
                        "age": 90,
                        # TODO NOINETINUDT
                        # "inet": ipaddress.ip_address("127.0.0.1"),
                        # TODO NOBLOBINUDT
                        # "blb": THE_BYTES,
                        "ts": THE_TIMESTAMP,
                    }
                )
            elif udt_format == "dataclass":
                udt_unit = ExtendedPlayer(
                    name="John",
                    age=90,
                    # TODO NOINETINUDT
                    # inet=ipaddress.ip_address("127.0.0.1"),
                    # TODO NOBLOBINUDT
                    # blb=THE_BYTES,
                    ts=THE_TIMESTAMP,
                )
            else:
                raise ValueError(f"Unknown udt format '{udt_format}'.")
        else:
            raise ValueError(f"Unknown udt mode '{udt_mode}'.")

        row_custom_colls = {
            "id": "row_custom_colls",
            "scalar_udt": udt_unit,
            "list_udt": [udt_unit],
            "set_udt": DataAPISet([udt_unit]),
            "map_udt": DataAPIMap([("k", udt_unit)]),
        }
        row_stdlib_colls = {
            "id": "row_stdlib_colls",
            "scalar_udt": udt_unit,
            "list_udt": [udt_unit],
            "set_udt": [udt_unit],
            "map_udt": {"k": udt_unit},
        }
        ins1c_result = await atable.insert_one(row_custom_colls)
        ins1s_result = await atable.insert_one(row_stdlib_colls)
        insm_result = await atable.insert_many([row_custom_colls, row_stdlib_colls])

        ins1_tuples = {ins1c_result.inserted_id_tuple, ins1s_result.inserted_id_tuple}
        assert ins1_tuples == set(insm_result.inserted_id_tuples)

        # read and verify, {custom, stdlib} x {deserializer in map, default behaviour}:

        # Custom datatypes, no deserializer specified:
        for custom_datatypes_in_reading in [True, False]:
            for deserializer_in_map in [False, True]:
                # prepare assets for read verification
                exp_row: dict[str, Any]
                expected_udt_unit: Any
                projection: dict[str, bool]

                deserializer_by_udt: dict[str, UDTDeserializerFunctionType]
                if deserializer_in_map:
                    if udt_mode == "simple":
                        deserializer_by_udt = {PLAYER_TYPE_NAME: _player_from_dict}
                    elif udt_mode in {"partial", "empty"}:
                        deserializer_by_udt = {
                            PLAYER_TYPE_NAME: _nullable_player_from_dict
                        }
                    elif udt_mode == "extended":
                        deserializer_by_udt = {
                            EXTENDED_PLAYER_TYPE_NAME: _extended_player_from_dict,
                        }
                    else:
                        raise ValueError(f"Unknown udt mode '{udt_mode}'.")
                else:
                    deserializer_by_udt = {}

                if custom_datatypes_in_reading:
                    projection = {"id": False}
                    if deserializer_in_map:
                        if udt_mode == "simple":
                            expected_udt_unit = Player(name="John", age=90)
                        elif udt_mode == "partial":
                            expected_udt_unit = NullablePlayer(age=90)
                        elif udt_mode == "empty":
                            expected_udt_unit = NullablePlayer()
                        elif udt_mode == "extended":
                            expected_udt_unit = ExtendedPlayer(
                                name="John",
                                age=90,
                                # TODO NOINETINUDT
                                # inet=ipaddress.ip_address("127.0.0.1"),
                                # TODO NOBLOBINUDT
                                # blb=THE_BYTES,
                                ts=THE_TIMESTAMP,
                            )
                        else:
                            raise ValueError(f"Unknown udt mode '{udt_mode}'.")
                    else:
                        if udt_mode == "simple":
                            expected_udt_unit = DataAPIDictUDT(
                                {
                                    "name": "John",
                                    "age": 90,
                                }
                            )
                        elif udt_mode == "partial":
                            expected_udt_unit = DataAPIDictUDT(
                                {"name": None, "age": 90}
                            )
                        elif udt_mode == "empty":
                            expected_udt_unit = DataAPIDictUDT(
                                {"name": None, "age": None}
                            )
                        elif udt_mode == "extended":
                            expected_udt_unit = DataAPIDictUDT(
                                {
                                    "name": "John",
                                    "age": 90,
                                    # TODO NOINETINUDT
                                    # "inet": ipaddress.ip_address("127.0.0.1"),
                                    # TODO NOBLOBINUDT
                                    # "blb": THE_BYTES,
                                    "ts": THE_TIMESTAMP,
                                }
                            )
                        else:
                            raise ValueError(f"Unknown udt mode '{udt_mode}'.")
                    exp_row = {
                        "scalar_udt": expected_udt_unit,
                        "list_udt": [expected_udt_unit],
                        "set_udt": DataAPISet([expected_udt_unit]),
                        "map_udt": DataAPIMap([("k", expected_udt_unit)]),
                    }
                else:
                    projection = {"id": False, "set_udt": False}
                    if deserializer_in_map:
                        if udt_mode == "simple":
                            expected_udt_unit = Player(name="John", age=90)
                        elif udt_mode == "partial":
                            expected_udt_unit = NullablePlayer(age=90)
                        elif udt_mode == "empty":
                            expected_udt_unit = NullablePlayer()
                        elif udt_mode == "extended":
                            expected_udt_unit = ExtendedPlayer(
                                name="John",
                                age=90,
                                # TODO NOINETINUDT
                                # inet=ipaddress.ip_address("127.0.0.1"),
                                # TODO NOBLOBINUDT
                                # blb=THE_BYTES,
                                ts=THE_DATETIME,
                            )
                        else:
                            raise ValueError(f"Unknown udt mode '{udt_mode}'.")
                    else:
                        if udt_mode == "simple":
                            expected_udt_unit = {
                                "name": "John",
                                "age": 90,
                            }
                        elif udt_mode == "partial":
                            expected_udt_unit = {"name": None, "age": 90}
                        elif udt_mode == "empty":
                            expected_udt_unit = {"name": None, "age": None}
                        elif udt_mode == "extended":
                            expected_udt_unit = {
                                "name": "John",
                                "age": 90,
                                # TODO NOINETINUDT
                                # "inet": ipaddress.ip_address("127.0.0.1"),
                                # TODO NOBLOBINUDT
                                # "blb": THE_BYTES,
                                "ts": THE_DATETIME,
                            }
                        else:
                            raise ValueError(f"Unknown udt mode '{udt_mode}'.")
                    exp_row = {
                        "scalar_udt": expected_udt_unit,
                        "list_udt": [expected_udt_unit],
                        "map_udt": {"k": expected_udt_unit},
                    }

                # table for reading
                reader_atable = atable.with_options(
                    api_options=APIOptions(
                        serdes_options=SerdesOptions(
                            custom_datatypes_in_reading=custom_datatypes_in_reading,
                            deserializer_by_udt=dict(deserializer_by_udt),
                            datetime_tzinfo=THE_TIMEZONE,
                        ),
                    ),
                )

                # read and check results
                row1 = await reader_atable.find_one(
                    {"id": "row_custom_colls"},
                    projection=projection,
                )
                row2 = await reader_atable.find_one(
                    {"id": "row_stdlib_colls"},
                    projection=projection,
                )
                dict_equal_same_class(row1, row2)
                dict_equal_same_class(row1, exp_row)

                # projection-based reads
                for fld in exp_row.keys() - projection.keys():
                    prj_read_row = await reader_atable.find_one(
                        {"id": "row_custom_colls"},
                        projection={fld: True},
                    )
                    dict_equal_same_class(prj_read_row, {fld: exp_row[fld]})

    @pytest.mark.describe("Test of null/partial UDT in DML, async")
    async def test_table_incomplete_udt_dml_async(
        self,
        player_udt: str,
        async_empty_table_udt_player: DefaultAsyncTable,
    ) -> None:
        atable = async_empty_table_udt_player.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True)
            ),
        )
        insert_results = await atable.insert_many(
            [
                {
                    "id": "nulls",
                    "scalar_udt": None,
                    "list_udt": [{}],
                    "set_udt": DataAPISet([{}]),
                    "map_udt": DataAPIMap({"k": {}}),
                },
                {
                    "id": "partials",
                    "scalar_udt": {"age": 75},
                    "list_udt": [{"age": 75}],
                    "set_udt": DataAPISet([{"age": 75}]),
                    "map_udt": DataAPIMap({"k": {"age": 75}}),
                },
            ]
        )
        assert insert_results.inserted_id_tuples == [("nulls",), ("partials",)]

        row_nulls = await atable.find_one({"id": "nulls"})
        null_udt_dict = DataAPIDictUDT({"name": None, "age": None})
        expected_row_nulls = {
            "id": "nulls",
            "scalar_udt": null_udt_dict,
            "list_udt": [null_udt_dict],
            "set_udt": DataAPISet([null_udt_dict]),
            "map_udt": DataAPIMap({"k": null_udt_dict}),
        }
        dict_equal_same_class(row_nulls, expected_row_nulls)

        row_partials = await atable.find_one({"id": "partials"})
        partial_udt_dict = DataAPIDictUDT({"name": None, "age": 75})
        expected_row_partials = {
            "id": "partials",
            "scalar_udt": partial_udt_dict,
            "list_udt": [partial_udt_dict],
            "set_udt": DataAPISet([partial_udt_dict]),
            "map_udt": DataAPIMap({"k": partial_udt_dict}),
        }
        dict_equal_same_class(row_partials, expected_row_partials)

    @pytest.mark.skipif(not CQL_AVAILABLE, reason="No CQL session available")
    @pytest.mark.describe("Test of weird UDT columns, async")
    async def test_table_udt_weirdcolumns_async(
        self,
        cql_session: Session,
        async_database: AsyncDatabase,
    ) -> None:
        try:
            for cql_statement in WEIRD_UDT_BASE_INITIALIZE_STATEMENTS:
                cql_session.execute(cql_statement)

            # test a read and a write for 'base weird'
            atable_weird_base = async_database.get_table(WEIRD_UDT_BASE_TABLE_NAME)
            ins_result = await atable_weird_base.insert_one(WEIRD_BASE_DOCUMENT)
            assert ins_result.inserted_id == WEIRD_BASE_DOCUMENT_PK
            doc_weird_base = await atable_weird_base.find_one(WEIRD_BASE_DOCUMENT_PK)
            assert doc_weird_base == WEIRD_BASE_DOCUMENT
        finally:
            for cql_statement in WEIRD_UDT_BASE_CLOSE_STATEMENTS:
                cql_session.execute(cql_statement)

    @pytest.mark.skipif(not CQL_AVAILABLE, reason="No CQL session available")
    @pytest.mark.describe("Test of unsupported UDT listing, async")
    async def test_table_udt_listunsupported_async(
        self,
        cql_session: Session,
        async_database: AsyncDatabase,
    ) -> None:
        try:
            cql_session.execute(UNSUPPORTED_UDT_CREATE)
            await asyncio.sleep(1.5)  # udt propagation requires some time, it seems

            assert UNSUPPORTED_UDT_NAME in await async_database.list_type_names()
            lt_result = await async_database.list_types()
            listed_udt_matches = [
                l_u
                for l_u in lt_result
                # sloppy matching
                if UNSUPPORTED_UDT_NAME in str(l_u.as_dict())
            ]
            assert len(listed_udt_matches) == 1
            listed_udt_match = listed_udt_matches[0]
            assert listed_udt_match.udt_type == TableUnsupportedColumnType.UNSUPPORTED
            assert listed_udt_match.api_support is not None
            assert UNSUPPORTED_UDT_NAME in listed_udt_match.api_support.cql_definition
        finally:
            cql_session.execute(UNSUPPORTED_UDT_DROP)

    @pytest.mark.describe("Test of UDT filtering, async")
    async def test_table_udt_filtering_async(
        self,
        async_empty_table_udtcollindexed: DefaultAsyncTable,
    ) -> None:
        atable = async_empty_table_udtcollindexed
        u_dict = DataAPIDictUDT({"name": "Otto", "age": 8})
        await atable.insert_one(
            {
                "id": "fullrow",
                "scalar_udt": u_dict,
                "set_udt": DataAPISet([u_dict]),
                "list_udt": [u_dict],
                "map_text_udt_e": {"ke": u_dict},
                "map_text_udt_v": {"kv": u_dict},
            }
        )

        # running different filtering patterns with UDTs
        x_dict = DataAPIDictUDT({"name": "Ada", "age": 8})
        prj = {"id": True}

        with pytest.raises(DataAPIResponseException):
            await atable.find_one(filter={"scalar_udt": u_dict})

        su_row0 = await atable.find_one(
            filter={"set_udt": {"$in": [u_dict, x_dict]}}, projection=prj
        )
        su_row1 = await atable.find_one(
            filter={"set_udt": {"$nin": [x_dict]}}, projection=prj
        )
        su_rown0 = await atable.find_one(
            filter={"set_udt": {"$in": [x_dict]}}, projection=prj
        )
        su_rown1 = await atable.find_one(
            filter={"set_udt": {"$nin": [x_dict, u_dict]}}, projection=prj
        )
        assert su_row0 is not None and su_row0["id"] == "fullrow"
        assert su_row1 is not None and su_row1["id"] == "fullrow"
        assert su_rown0 is None
        assert su_rown1 is None

        lu_row0 = await atable.find_one(
            filter={"list_udt": {"$in": [u_dict, x_dict]}}, projection=prj
        )
        lu_row1 = await atable.find_one(
            filter={"list_udt": {"$nin": [x_dict]}}, projection=prj
        )
        lu_rown0 = await atable.find_one(
            filter={"list_udt": {"$in": [x_dict]}}, projection=prj
        )
        lu_rown1 = await atable.find_one(
            filter={"list_udt": {"$nin": [x_dict, u_dict]}}, projection=prj
        )
        assert lu_row0 is not None and lu_row0["id"] == "fullrow"
        assert lu_row1 is not None and lu_row1["id"] == "fullrow"
        assert lu_rown0 is None
        assert lu_rown1 is None

        meu_row0 = await atable.find_one(
            filter={"map_text_udt_e": {"$in": [["ke", u_dict], ["ke", x_dict]]}},
            projection=prj,
        )
        meu_row1 = await atable.find_one(
            filter={"map_text_udt_e": {"$nin": [["ke", x_dict]]}}, projection=prj
        )
        meu_rown0 = await atable.find_one(
            filter={"map_text_udt_e": {"$in": [["KE", u_dict], ["ke", x_dict]]}},
            projection=prj,
        )
        meu_rown1 = await atable.find_one(
            filter={"map_text_udt_e": {"$nin": [["ke", u_dict], ["ke", x_dict]]}},
            projection=prj,
        )
        assert meu_row0 is not None and meu_row0["id"] == "fullrow"
        assert meu_row1 is not None and meu_row1["id"] == "fullrow"
        assert meu_rown0 is None
        assert meu_rown1 is None

        mvu_row0 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$in": [u_dict, x_dict]}}},
            projection=prj,
        )
        mvu_row1 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$nin": [x_dict]}}},
            projection=prj,
        )
        mvu_row2 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$all": [u_dict]}}},
            projection=prj,
        )
        mvu_rown0 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$in": [x_dict]}}},
            projection=prj,
        )
        mvu_rown1 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$nin": [u_dict, x_dict]}}},
            projection=prj,
        )
        mvu_rown2 = await atable.find_one(
            filter={"map_text_udt_v": {"$values": {"$all": [u_dict, x_dict]}}},
            projection=prj,
        )
        assert mvu_row0 is not None and mvu_row0["id"] == "fullrow"
        assert mvu_row1 is not None and mvu_row1["id"] == "fullrow"
        assert mvu_row2 is not None and mvu_row2["id"] == "fullrow"
        assert mvu_rown0 is None
        assert mvu_rown1 is None
        assert mvu_rown2 is None

        # using partial UDTs for filtering in various ways
        p_dict = DataAPIDictUDT({"age": 8})

        with pytest.raises(DataAPIResponseException):
            await atable.find_one(filter={"scalar_udt": p_dict})

        psu_rown0 = await atable.find_one(filter={"set_udt": {"$in": [p_dict]}})
        assert psu_rown0 is None
        plu_rown0 = await atable.find_one(filter={"list_udt": {"$in": [p_dict]}})
        assert plu_rown0 is None
        pmeu_rown0 = await atable.find_one(
            filter={"map_text_udt_e": {"$in": [["h", p_dict]]}}
        )
        assert pmeu_rown0 is None
        pmvu_rown0 = await atable.find_one(
            filter={"map_text_udt_e": {"$values": {"$in": [p_dict]}}}
        )
        assert pmvu_rown0 is None
        pmvu_rown1 = await atable.find_one(
            filter={"map_text_udt_e": {"$values": {"$all": [p_dict]}}}
        )
        assert pmvu_rown1 is None

        await atable.insert_one(
            {
                "id": "withnulls",
                "scalar_udt": p_dict,
                "set_udt": DataAPISet([p_dict]),
                "list_udt": [p_dict],
                "map_text_udt_e": {"he": p_dict},
                "map_text_udt_v": {"hv": p_dict},
            }
        )

        psu_row0 = await atable.find_one(
            filter={"set_udt": {"$in": [p_dict]}}, projection=prj
        )
        assert psu_row0 is not None and psu_row0["id"] == "withnulls"
        plu_row0 = await atable.find_one(
            filter={"list_udt": {"$in": [p_dict]}}, projection=prj
        )
        assert plu_row0 is not None and plu_row0["id"] == "withnulls"
        pmeu_row0 = await atable.find_one(
            filter={"map_text_udt_e": {"$in": [["he", p_dict]]}}, projection=prj
        )
        assert pmeu_row0 is not None and pmeu_row0["id"] == "withnulls"
        pmvu_row0 = await atable.find_one(
            filter={"map_text_udt_e": {"$values": {"$in": [p_dict]}}}, projection=prj
        )
        assert pmvu_row0 is not None and pmvu_row0["id"] == "withnulls"
        pmvu_row1 = await atable.find_one(
            filter={"map_text_udt_e": {"$values": {"$all": [p_dict]}}}, projection=prj
        )
        assert pmvu_row1 is not None and pmvu_row1["id"] == "withnulls"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any

from astrapy.event_observers.events import ObservableEvent, ObservableEventType


class Observer(ABC):
    """
    An observer that can be attached to astrapy events through the API options.

    Users can subclass Observer and provide their implementation
    of the `receive` method. Request-issuing classes (such as Database or Table)
    will dispatch events to the observers registered in their API options.

    This class offers two static factory methods for common use-cases:
    `from_event_list` and `from_event_dict`.
    """

    enabled: bool = True

    @abstractmethod
    def receive(
        self,
        event: ObservableEvent,
        sender: Any = None,
        function_name: str | None = None,
        request_id: str | None = None,
    ) -> None:
        """Receive and event.

        Args:
            event: the event that astrapy is dispatching to the observer.
            sender: the object directly responsible for generating the event.
            function_name: when applicable, the name of the function/method
                that triggered the event.
            request_id: an optional ID used to group different received events
                as occurring within the lifecycle of a single HTTP request.
        """
        ...

    @staticmethod
    def from_event_list(
        event_list: list[ObservableEvent],
        *,
        event_types: Iterable[ObservableEventType] | None = None,
    ) -> Observer:
        """
        Create an Observer object wrapping a caller-provided list.

        The resulting observer will simply append the events it receives into
        the list.

        Args:
            event_list: the list where the caller will find the received events.
            event_types: if provided, it's a list of event types so that only
                events matching this filter are processed.

        Example:
            >>> from astrapy.api_options import APIOptions
            >>> from astrapy.event_observers import ObservableEvent, Observer
            >>> my_ev_list: list[ObservableEvent] = []
            >>> my_observer = Observer.from_event_list(my_ev_list)
            >>> instrumented_table = my_table.with_options(
            ...     api_options=APIOptions(
            ...         event_observers={"obs000": my_observer},
            ...     ),
            ... )
            >>> # start using 'instrumented_table' ...
            >>> # ... and then inspect 'my_ev_list'
        """

        class _ObserverFromList(Observer):
            def __init__(
                self,
                _event_list: list[ObservableEvent],
                _event_types: Iterable[ObservableEventType] | None,
            ) -> None:
                self.event_list = _event_list
                self.event_types = (
                    set(ObservableEventType.__members__.values())
                    if _event_types is None
                    else set(_event_types)
                )

            def receive(
                self,
                event: ObservableEvent,
                sender: Any = None,
                function_name: str | None = None,
                request_id: str | None = None,
            ) -> None:
                if event.event_type in self.event_types:
                    self.event_list.append(event)

        return _ObserverFromList(event_list, event_types)

    @staticmethod
    def from_event_dict(
        event_dict: dict[ObservableEventType, list[ObservableEvent]],
        *,
        event_types: Iterable[ObservableEventType] | None = None,
    ) -> Observer:
        """
        Create an Observer object wrapping a caller-provided dictionary.

        The resulting observer will simply append the events it receives into
        the dictionary, grouped by event type. Dict values are lists of events.

        Args:
            event_dict: the dict where the caller will find the received events.
            event_types: if provided, it's a list of event types so that only
                events matching this filter are processed.

        Example:
            >>> from astrapy.api_options import APIOptions
            >>> from astrapy.event_observers import (
            ...     ObservableEvent,
            ...     ObservableEventType,
            ...     Observer,
            ... )
            >>> my_ev_map: dict[ObservableEventType, list[ObservableEvent]] = {}
            >>> my_observer = Observer.from_event_dict(my_ev_map)
            >>> instrumented_table = my_table.with_options(
            ...     api_options=APIOptions(
            ...         event_observers={"obs000": my_observer},
            ...     ),
            ... )
            >>> # start using 'instrumented_table' ...
            >>> # ... and then inspect 'my_ev_map'
        """

        class _ObserverFromDict(Observer):
            def __init__(
                self,
                _event_dict: dict[ObservableEventType, list[ObservableEvent]],
                _event_types: Iterable[ObservableEventType] | None,
            ) -> None:
                self.event_dict = _event_dict
                self.event_types = (
                    set(ObservableEventType.__members__.values())
                    if _event_types is None
                    else set(_event_types)
                )

            def receive(
                self,
                event: ObservableEvent,
                sender: Any = None,
                function_name: str | None = None,
                request_id: str | None = None,
            ) -> None:
                if event.event_type in self.event_types:
                    self.event_dict[event.event_type] = self.event_dict.get(
                        event.event_type, []
                    ) + [event]

        return _ObserverFromDict(event_dict, event_types)
